#
#    Copyright 2017 Mirantis Inc.
#
#    Copyright (c) 2022 Dell Inc. or its subsidiaries.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ipaddress
import random

from oslo_utils import uuidutils
from tempest import config
from tempest.lib.common.utils.linux import remote_client
from tempest.lib.common.utils import test_utils
from tempest.lib import exceptions as lib_exc
from tempest.scenario import manager

from ironic_tempest_plugin.services.baremetal import base
from ironic_tempest_plugin.tests.scenario import baremetal_manager as bm

CONF = config.CONF


class BaremetalStandaloneManager(bm.BaremetalScenarioTest,
                                 manager.NetworkScenarioTest):

    credentials = ['primary', 'admin', 'system_admin']
    # NOTE(vsaienko): Standalone tests are using v1/node/<node_ident>/vifs to
    # attach VIF to a node.
    min_microversion = '1.28'

    image_ref = None
    image_checksum = None
    boot_option = None

    @classmethod
    def skip_checks(cls):
        """Defines conditions to skip these tests."""
        super(BaremetalStandaloneManager, cls).skip_checks()
        if CONF.service_available.nova:
            raise cls.skipException('Nova is enabled. Stand-alone tests will '
                                    'be skipped.')

    @classmethod
    def create_networks(cls):
        """Create a network with a subnet connected to a router.

        Return existed network specified in compute/fixed_network_name
        config option.
        TODO(vsaienko): Add network/subnet/router when we setup
        ironic-standalone with multitenancy.

        :returns: network, subnet, router
        """
        network = None
        subnet = None
        router = None
        if CONF.network.shared_physical_network:
            if not CONF.compute.fixed_network_name:
                m = ('Configuration option "[compute]/fixed_network_name" '
                     'must be set.')
                raise lib_exc.InvalidConfiguration(m)
            network = cls.os_admin.networks_client.list_networks(
                name=CONF.compute.fixed_network_name)['networks'][0]
        return network, subnet, router

    @classmethod
    def get_available_nodes(cls):
        """Get all ironic nodes that can be deployed.

        We can deploy on nodes when the following conditions are met:
          * provision_state is 'available'
          * maintenance is False
          * No instance_uuid is associated to node.

        :returns: a list of Ironic nodes.
        """
        fields = ['uuid', 'driver', 'instance_uuid', 'provision_state',
                  'name', 'maintenance']
        _, body = cls.baremetal_client.list_nodes(provision_state='available',
                                                  associated=False,
                                                  maintenance=False,
                                                  fields=','.join(fields))
        return body['nodes']

    @classmethod
    def get_random_available_node(cls):
        """Randomly pick an available node for deployment."""
        nodes = cls.get_available_nodes()
        if nodes:
            return random.choice(nodes)

    @classmethod
    def create_neutron_port(cls, *args, **kwargs):
        """Creates a neutron port.

        For a full list of available parameters, please refer to the official
        API reference:
        https://docs.openstack.org/api-ref/network/v2/index.html#create-port

        :returns: server response body.
        """
        port = cls.ports_client.create_port(*args, **kwargs)['port']
        return port

    @classmethod
    def _associate_instance_with_node(cls, node_id, instance_uuid):
        """Update instance_uuid for a given node.

        :param node_id: Name or UUID of the node.
        :param instance_uuid: UUID of the instance to associate.
        :returns: server response body.
        """
        _, body = cls.baremetal_client.update_node(
            node_id, instance_uuid=instance_uuid)
        return body

    @classmethod
    def _disassociate_instance_with_node(cls, node_id):
        """Disassociate instance_uuid from attached node.

        :param node_id: Name or UUID of the node.
        """
        cls.update_node(node_id, [{'op': 'replace',
                                   'path': '/instance_uuid',
                                   'value': None}])

    @classmethod
    def get_node_vifs(cls, node_id):
        """Return a list of VIFs for a given node.

        :param node_id: Name or UUID of the node.
        :returns: A list of VIFs associated with the node.
        """
        _, body = cls.baremetal_client.vif_list(node_id)
        vifs = [v['id'] for v in body['vifs']]
        return vifs

    @classmethod
    def add_floatingip_to_node(cls, node_id):
        """Add floating IP to node.

        Create and associate floating IP with node VIF.

        :param node_id: Name or UUID of the node.
        :returns: IP address of associated floating IP.
        """
        vif = cls.get_node_vifs(node_id)[0]
        body = cls.floating_ips_client.create_floatingip(
            floating_network_id=CONF.network.public_network_id)
        floating_ip = body['floatingip']
        cls.floating_ips_client.update_floatingip(floating_ip['id'],
                                                  port_id=vif)
        return floating_ip['floating_ip_address']

    @classmethod
    def get_server_ip(cls, node_id):
        """Get the server fixed IP.

        :param node_id: Name or UUID of the node.
        :returns: IP address of associated fixed IP.
        """
        vif = cls.get_node_vifs(node_id)[0]
        body = cls.ports_client.show_port(vif)['port']
        fixed_ip = body['fixed_ips'][0]
        return fixed_ip['ip_address']

    @classmethod
    def cleanup_floating_ip(cls, ip_address):
        """Removes floating IP."""
        body = cls.os_admin.floating_ips_client.list_floatingips()
        floating_ip_id = [f['id'] for f in body['floatingips'] if
                          f['floating_ip_address'] == ip_address][0]
        cls.os_admin.floating_ips_client.delete_floatingip(floating_ip_id)

    @classmethod
    @bm.retry_on_conflict
    def detach_all_vifs_from_node(cls, node_id, force_delete=False):
        """Detach all VIFs from a given node.

        :param node_id: Name or UUID of the node.
        """
        vifs = cls.get_node_vifs(node_id)
        for vif in vifs:
            try:
                cls.baremetal_client.vif_detach(node_id, vif)
            except lib_exc.BadRequest:
                # When the vif was already removed, such as the
                # node was already unprovisioned.
                pass
            if force_delete:
                try:
                    cls.ports_client.delete_port(vif)
                except lib_exc.NotFound:
                    pass

    @classmethod
    @bm.retry_on_conflict
    def vif_attach(cls, node_id, vif_id):
        """Attach VIF to a give node.

        :param node_id: Name or UUID of the node.
        :param vif_id: Identifier of the VIF to attach.
        """
        cls.baremetal_client.vif_attach(node_id, vif_id)

    @classmethod
    def get_and_reserve_node(cls, node=None):
        """Pick an available node for deployment and reserve it.

        Only one instance_uuid may be associated, use this behaviour as
        reservation node when tests are launched concurrently. If node is
        not passed directly pick random available for deployment node.

        :param node: Ironic node to associate instance_uuid with.
        :returns: Ironic node.
        """
        instance_uuid = uuidutils.generate_uuid()
        nodes = []

        def _try_to_associate_instance():
            n = node or cls.get_random_available_node()
            if n is None:
                return False
            try:
                cls._associate_instance_with_node(n['uuid'], instance_uuid)
                nodes.append(n)
            except lib_exc.Conflict:
                return False
            return True

        if (not test_utils.call_until_true(
                _try_to_associate_instance,
                duration=CONF.baremetal.association_timeout, sleep_for=1)):
            msg = ('Timed out waiting to associate instance to ironic node '
                   'uuid %s' % instance_uuid)
            raise lib_exc.TimeoutException(msg)

        return nodes[0]

    @classmethod
    def unreserve_node(cls, node):
        """Unreserves node by disassociating instance_uuid attached to node.

        :param node: Ironic node to disassociate instance_uuid from.
        """

        def _try_to_disassociate_instance():
            _, node_prop = cls.baremetal_client.show_node(node['uuid'])
            if node_prop['instance_uuid']:
                try:
                    cls._disassociate_instance_with_node(node['uuid'])
                except lib_exc.Conflict:
                    return False
            return True
        if (not test_utils.call_until_true(
                _try_to_disassociate_instance,
                duration=CONF.baremetal.association_timeout, sleep_for=1)):
            msg = ('Timed out waiting to disassociate instance from '
                   'ironic node uuid %s' % node['instance_uuid'])
            raise lib_exc.TimeoutException(msg)

    @classmethod
    def gen_config_drive_net_info(cls, node_id, n_port):
        # Find the port with the vif.
        use_port = None
        _, body = cls.baremetal_client.list_node_ports(node_id)
        for port in body['ports']:
            _, p = cls.baremetal_client.show_port(port['uuid'])
            if 'tenant_vif_port_id' in p['internal_info']:
                use_port = p
                break
        if not use_port:
            m = ('Unable to determine proper mac address to use for config '
                 'to apply for the virtual media port test.')
            raise lib_exc.InvalidConfiguration(m)
        vif_mac_address = use_port['address']
        if CONF.validation.ip_version_for_ssh == 4:
            ip_version = "ipv4"
        else:
            ip_version = "ipv6"
        ip_address = n_port['fixed_ips'][0]['ip_address']
        subnet_id = n_port['fixed_ips'][0]['subnet_id']
        subnet = cls.os_primary.subnets_client.show_subnet(
            subnet_id).get('subnet')
        ip_netmask = str(ipaddress.ip_network(subnet.get('cidr')).netmask)
        if ip_version == "ipv4":
            route = [{
                "netmask": "0.0.0.0",
                "network": "0.0.0.0",
                "gateway": subnet.get('gateway_ip'),
            }]
        else:
            # Eh... the data structure doesn't really allow for
            # this to be easy since a default route with v6
            # is just referred to as ::/0
            # so network and netmask would be ::, which is
            # semi-mind-breaking. Anyway, route advertisers are
            # expected in this case.
            route = []

        return {
            "links": [{"id": "port-test",
                       "type": "vif",
                       "ethernet_mac_address": vif_mac_address}],
            "networks": [
                {
                    "id": "network0",
                    "type": ip_version,
                    "link": "port-test",
                    "ip_address": ip_address,
                    "netmask": ip_netmask,
                    "network_id": "network0",
                    "routes": route
                }
            ],
            "services": []
        }

    @classmethod
    def boot_node(cls, image_ref=None, image_checksum=None,
                  boot_option=None, config_drive_networking=False,
                  fallback_network=None):
        """Boot ironic node.

        The following actions are executed:
          * Create/Pick networks to boot node in.
          * Create Neutron port and attach it to node.
          * Update node image_source/root_gb.
          * Deploy node.
          * Wait until node is deployed.

        :param image_ref: Reference to user image to boot node with.
        :param image_checksum: md5sum of image specified in image_ref.
                               Needed only when direct HTTP link is provided.
        :param boot_option: The default boot option to utilize. If not
                            specified, the ironic deployment default shall
                            be utilized.
        :param config_drive_networking: If we should load configuration drive
            with network_data values.
        :param fallback_network: Network to use if we are not able to detect
            a network for use.
        """
        config_drive = {}

        if image_ref is None:
            image_ref = cls.image_ref
        if image_checksum is None:
            image_checksum = cls.image_checksum
        if boot_option is None:
            boot_option = cls.boot_option

        network, subnet, router = cls.create_networks()
        try:
            n_port = cls.create_neutron_port(network_id=network['id'])

        except TypeError:
            if fallback_network:
                n_port = cls.create_neutron_port(
                    network_id=fallback_network)
            else:
                raise
        cls.vif_attach(node_id=cls.node['uuid'], vif_id=n_port['id'])
        config_drive = None
        if config_drive_networking:
            config_drive = {}
            config_drive['network_data'] = cls.gen_config_drive_net_info(
                cls.node['uuid'], n_port)

        patch = [{'path': '/instance_info/image_source',
                  'op': 'add',
                  'value': image_ref}]
        if image_checksum is not None:
            patch.append({'path': '/instance_info/image_checksum',
                          'op': 'add',
                          'value': image_checksum})
        patch.append({'path': '/instance_info/root_gb',
                      'op': 'add',
                      'value': CONF.baremetal.adjusted_root_disk_size_gb})

        if boot_option:
            patch.append({'path': '/instance_info/capabilities',
                          'op': 'add',
                          'value': {'boot_option': boot_option}})
        cls.update_node(cls.node['uuid'], patch=patch)

        if not config_drive:
            cls.set_node_provision_state(cls.node['uuid'], 'active')
        else:
            cls.set_node_provision_state(
                cls.node['uuid'], 'active',
                configdrive=config_drive)

        cls.wait_power_state(cls.node['uuid'],
                             bm.BaremetalPowerStates.POWER_ON)
        cls.wait_provisioning_state(cls.node['uuid'],
                                    bm.BaremetalProvisionStates.ACTIVE,
                                    timeout=CONF.baremetal.active_timeout,
                                    interval=30)

    @classmethod
    def terminate_node(cls, node_id, force_delete=False):
        """Terminate active ironic node.

        The following actions are executed:
           * Detach all VIFs from the given node.
           * Unprovision node.
           * Wait until node become available.

        :param node_id: Name or UUID for the node.
        """
        cls.detach_all_vifs_from_node(node_id, force_delete=force_delete)

        if cls.delete_node or force_delete:
            node_state = cls.get_node(node_id)['provision_state']
            if node_state != bm.BaremetalProvisionStates.AVAILABLE:
                # Check the state before making the call, to permit tests to
                # drive node into a clean state before exiting the test, which
                # is needed for some tests because of complex tests.
                cls.set_node_provision_state(node_id, 'deleted')
            # NOTE(vsaienko) We expect here fast switching from deleted to
            # available as automated cleaning is disabled so poll status
            # each 1s.
            cls.wait_provisioning_state(
                node_id,
                [bm.BaremetalProvisionStates.NOSTATE,
                 bm.BaremetalProvisionStates.AVAILABLE],
                timeout=CONF.baremetal.unprovision_timeout,
                interval=1)

    @classmethod
    def rescue_node(cls, node_id, rescue_password):
        """Rescue the node."""
        cls.set_node_provision_state(node_id, 'rescue',
                                     rescue_password=rescue_password)
        cls.wait_provisioning_state(
            node_id,
            bm.BaremetalProvisionStates.RESCUE,
            timeout=CONF.baremetal.rescue_timeout,
            interval=1)

    @classmethod
    def unrescue_node(cls, node_id):
        """Unrescue the node."""
        cls.set_node_provision_state(node_id, 'unrescue')
        cls.wait_provisioning_state(
            node_id,
            bm.BaremetalProvisionStates.ACTIVE,
            timeout=CONF.baremetal.unrescue_timeout,
            interval=1)

    def manual_cleaning(self, node, clean_steps=None, runbook=None):
        """Performs manual cleaning.

        The following actions are executed:
           * Expects node to be in available state.
           * Brings the node to manageable state.
           * Do manual cleaning.
           * Brings the node back to original available.

        :param node: Ironic node to associate instance_uuid with.
        :param clean_steps: clean steps for manual cleaning.
        :param runbook: unique identifier of a runbook.
        """
        if clean_steps is None and runbook is None:
            raise ValueError("Either clean_steps or runbook must be provided.")

        self.set_node_provision_state(node['uuid'], 'manage')
        self.wait_provisioning_state(
            node['uuid'],
            [bm.BaremetalProvisionStates.MANAGEABLE],
            timeout=CONF.baremetal.unprovision_timeout,
            interval=30)

        if runbook:
            self.set_node_provision_state(
                node['uuid'], 'clean', runbook=runbook)
        else:
            self.set_node_provision_state(
                node['uuid'], 'clean', clean_steps=clean_steps)

        self.wait_provisioning_state(
            node['uuid'],
            [bm.BaremetalProvisionStates.MANAGEABLE],
            timeout=CONF.baremetal.unprovision_timeout,
            interval=30)
        self.set_node_provision_state(node['uuid'], 'provide')
        self.wait_provisioning_state(
            node['uuid'],
            [bm.BaremetalProvisionStates.NOSTATE,
             bm.BaremetalProvisionStates.AVAILABLE],
            timeout=CONF.baremetal.unprovision_timeout,
            interval=30)

    def manual_cleaning_with_runbook(self, node):
        steps = [{
            'interface': 'bios',
            'step': 'apply_configuration',
            'args': {},
            'order': 1
        }]
        _, runbook = self.baremetal_client.create_runbook('CUSTOM_AWESOME',
                                                          steps=steps)
        self.manual_cleaning(node, runbook=runbook)

    def check_manual_partition_cleaning(self, node):
        """Tests the cleanup step for erasing devices metadata.

        :param node: Ironic node to associate instance_uuid with, it is
            expected to be in 'active' state
        """
        clean_steps = [
            {
                "interface": "deploy",
                "step": "erase_devices_metadata"
            }
        ]
        self.manual_cleaning(node, clean_steps=clean_steps)
        # TODO(yolanda): we currently are not checking it the cleanup
        # was actually removing the metadata, because there was not a good
        # way to achieve that check for vms and baremetal

    def check_bios_apply_and_reset_configuration(self, node, settings):
        clean_steps = [
            {
                "interface": "bios",
                "step": "apply_configuration",
                "args": {"settings": settings}
            }
        ]
        self.manual_cleaning(node, clean_steps=clean_steps)

        # query the api to check node bios settings
        _, bios_settings = self.baremetal_client.list_node_bios_settings(
            node['uuid'])

        for setting in settings:
            self.assertIn(setting['name'],
                          [i['name'] for i in bios_settings['bios']])
            self.assertIn(setting['value'],
                          [i['value'] for i in bios_settings['bios']])

        # reset bios and ensure that the settings are not there
        clean_steps = [
            {
                "interface": "bios",
                "step": "factory_reset"
            }
        ]
        self.manual_cleaning(node, clean_steps=clean_steps)
        _, bios_settings = self.baremetal_client.list_node_bios_settings(
            node['uuid'])
        self.assertEqual([], bios_settings['bios'])


class BaremetalStandaloneScenarioTest(BaremetalStandaloneManager):

    # API microversion to use among all calls
    api_microversion = '1.28'

    # The node driver to use in the test
    driver = None

    # If we do not require an explicit driver to be set as a base for
    # the test. This is useful for tests where we perform a vertical
    # slice of drivers to test as opposed to focusing all testing
    # around the driver itself. This allows for us to verify the
    # driver interfaces align with the established contracts,
    # and allows the same test to use ipmi, redfish, or some other
    # known driver to work.
    use_available_driver = None

    # If we don't require an explicit driver, then what drivers *can* we
    # operate with. In essence, this exists to prevent the test from failing
    # on 3rd party drivers, and vendor specific driers which do not support
    # the sort of interfaces we may be trying to test by default.
    valid_driver_list = []

    # The bios interface to use by the HW type. The bios interface of the
    # node used in the test will be set to this value. If set to None, the
    # node will retain its existing bios_interface value (which may have been
    # set via a different test).
    bios_interface = None

    # The deploy interface to use by the HW type. The deploy interface of
    # the node used in the test will be set to this value. If set to None,
    # the node will retain its existing deploy_interface value (which may have
    # been set via a different test).
    deploy_interface = None

    # The rescue interface to use by the HW type. The rescue interface of
    # the node used in the test will be set to this value. If set to None,
    # the node will retain its existing rescue_interface value (which may have
    # been set via a different test).
    rescue_interface = None

    # The boot interface to use by the HW type. The boot interface of the
    # node used in the test will be set to this value. If set to None, the
    # node will retain its existing boot_interface value (which may have been
    # set via a different test).
    boot_interface = None

    # The raid interface to use by the HW type. The raid interface of the
    # node used in the test will be set to this value. If set to None, the
    # node will retain its existing raid_interface value (which may have been
    # set via a different test).
    raid_interface = None

    # The management interface to use by the HW type. The management interface
    # of the node used in the test will be set to this value. If set to None,
    # the node will retain its existing management_interface value (which may
    # have been set via a different test).
    management_interface = None

    # The power interface to use by the HW type. The power interface of the
    # node used in the test will be set to this value. If set to None, the
    # node will retain its existing power_interface value (which may have been
    # set via a different test).
    power_interface = None

    # The inspect interface to use by the HW type. The inspect interface of the
    # node used in the test will be set to this value. If set to None, the
    # node will retain its existing inspect_interface value (which may have
    # been set via a different test).
    inspect_interface = None

    # Boolean value specify if image is wholedisk or not.
    wholedisk_image = None

    # If we need to set provision state 'deleted' for the node  after test
    delete_node = True

    mandatory_attr = ['driver', 'image_ref']

    node = None
    node_ip = None

    @classmethod
    def skip_checks(cls):
        super(BaremetalStandaloneScenarioTest, cls).skip_checks()
        if (not cls.use_available_driver
            and cls.driver not in CONF.baremetal.enabled_drivers
                + CONF.baremetal.enabled_hardware_types):
            raise cls.skipException(
                'The driver: %(driver)s used in test is not in the list of '
                'enabled_drivers %(enabled_drivers)s or '
                'enabled_hardware_types %(enabled_hw_types)s '
                'in the tempest config.' % {
                    'driver': cls.driver,
                    'enabled_drivers': CONF.baremetal.enabled_drivers,
                    'enabled_hw_types': CONF.baremetal.enabled_hardware_types})
        for iface in base.SUPPORTED_INTERFACES:
            requested = getattr(cls, f'{iface}_interface')
            enabled = getattr(CONF.baremetal, f'enabled_{iface}_interfaces')
            if requested and requested not in enabled:
                raise cls.skipException(
                    "%(type)s interface %(iface)s required by the test is not "
                    "in the list of enabled %(type)s interfaces "
                    "%(enabled)s" % {
                        'iface': requested,
                        'type': iface,
                        'enabled': ', '.join(enabled),
                    })
        if (cls.wholedisk_image is not None
                and not cls.wholedisk_image
                and CONF.baremetal.use_provision_network):
            # We only want to enter here if cls.wholedisk_image is set to
            # a value. If None, skip, if True/False go from there.
            raise cls.skipException(
                'Partitioned images are not supported with multitenancy.')

    @classmethod
    def set_node_to_active(cls, image_ref=None, image_checksum=None,
                           fallback_network=None,
                           config_drive_networking=None,
                           method_to_get_ip=None):
        cls.boot_node(image_ref, image_checksum,
                      fallback_network=fallback_network,
                      config_drive_networking=config_drive_networking)
        if method_to_get_ip:
            cls.node_ip = method_to_get_ip(cls.node['uuid'])
        elif CONF.validation.connect_method == 'floating':
            cls.node_ip = cls.add_floatingip_to_node(cls.node['uuid'])
        elif CONF.validation.connect_method == 'fixed':
            cls.node_ip = cls.get_server_ip(cls.node['uuid'])
        else:
            m = ('Configuration option "[validation]/connect_method" '
                 'must be set.')
            raise lib_exc.InvalidConfiguration(m)

    @classmethod
    def resource_setup(cls):
        super(BaremetalStandaloneScenarioTest, cls).resource_setup()
        base.set_baremetal_api_microversion(cls.api_microversion)
        for v in cls.mandatory_attr:
            if getattr(cls, v) is None:
                raise lib_exc.InvalidConfiguration(
                    "Mandatory attribute %s not set." % v)
        image_checksum = None
        if not uuidutils.is_uuid_like(cls.image_ref):
            image_checksum = cls.image_checksum
        boot_kwargs = {'image_checksum': image_checksum}
        for iface in base.SUPPORTED_INTERFACES:
            if requested := getattr(cls, f'{iface}_interface'):
                boot_kwargs[f'{iface}_interface'] = requested

        # just get an available node
        cls.node = cls.get_and_reserve_node()
        if (cls.use_available_driver
                and not cls.driver
                and cls.node['driver'] in cls.valid_driver_list):
            # If we're attempting to reuse the existing driver, then
            # lets save a value for update_node_driver to work with.
            cls.driver = cls.node['driver']
        cls.update_node_driver(cls.node['uuid'], cls.driver, **boot_kwargs)

    @classmethod
    def cleanup_vif_attachments(cls):
        vifs = cls.get_node_vifs(cls.node['uuid'])
        # Remove ports before deleting node, to catch regression for cases
        # when user did this prior unprovision node.
        for vif in vifs:
            try:
                cls.ports_client.delete_port(vif)
            except lib_exc.NotFound:
                pass

    @classmethod
    def resource_cleanup(cls):
        if CONF.validation.connect_method == 'floating':
            if cls.node_ip:
                try:
                    cls.cleanup_floating_ip(cls.node_ip)
                except IndexError:
                    # There is no fip to actually remove in this case.
                    pass

        cls.cleanup_vif_attachments()
        cls.terminate_node(cls.node['uuid'])
        cls.unreserve_node(cls.node)
        base.reset_baremetal_api_microversion()
        super(BaremetalStandaloneManager, cls).resource_cleanup()

    def boot_and_verify_node(self, image_ref=None, image_checksum=None,
                             should_succeed=True):
        self.set_node_to_active(image_ref, image_checksum)
        self.assertTrue(self.ping_ip_address(self.node_ip,
                                             should_succeed=should_succeed))

    def build_raid_and_verify_node(self, config=None, deploy_time=False,
                                   erase_device_metadata=True):
        config = config or self.raid_config
        if deploy_time:
            steps = [
                {
                    "interface": "deploy",
                    "step": "erase_devices_metadata",
                    "priority": 98,
                    "args": {},
                },
                {
                    "interface": "raid",
                    "step": "apply_configuration",
                    "priority": 97,
                    "args": {"raid_config": config},
                }
            ]
            _, template = self.baremetal_client.create_deploy_template(
                'CUSTOM_RAID', steps=steps)
            # Set a cleanup to ensure the deploy template is removed.
            self.addCleanup(
                self.baremetal_client.delete_deploy_template,
                template['uuid'])
            self.baremetal_client.add_node_trait(self.node['uuid'],
                                                 'CUSTOM_RAID')

        else:
            steps = [
                {
                    "interface": "raid",
                    "step": "delete_configuration"
                },
                {
                    "interface": "raid",
                    "step": "create_configuration",
                }
            ]
            if erase_device_metadata:
                steps.insert(1, {
                    "interface": "deploy",
                    "step": "erase_devices_metadata"})
            self.baremetal_client.set_node_raid_config(self.node['uuid'],
                                                       config)
            self.manual_cleaning(self.node, clean_steps=steps)

        # The node has been changed, anything at this point, we need to back
        # out the raid configuration.
        if not deploy_time:
            self.addCleanup(self.remove_raid_configuration, self.node,
                            erase_device_metadata=erase_device_metadata)

        # NOTE(dtantsur): this is not required, but it allows us to check that
        # the RAID device was in fact created and is used for deployment.
        patch = [{'path': '/properties/root_device',
                  'op': 'add', 'value': {
                      'name': CONF.baremetal.root_device_name}}]
        if deploy_time:
            patch.append({'path': '/instance_info/traits',
                          'op': 'add', 'value': ['CUSTOM_RAID']})
        self.update_node(self.node['uuid'], patch=patch)
        # NOTE(dtantsur): apparently cirros cannot boot from md devices :(
        # So we only move the node to active (verifying deployment).
        self.set_node_to_active()
        if deploy_time:
            self.addCleanup(self.remove_raid_configuration, self.node)

    def remove_root_device_hint(self):
        patch = [{'path': '/properties/root_device',
                  'op': 'remove'}]
        self.update_node(self.node['uuid'], patch=patch)

    def remove_raid_configuration(self, node, erase_device_metadata=True):
        self.baremetal_client.set_node_raid_config(node['uuid'], {})
        steps = [
            {
                "interface": "raid",
                "step": "delete_configuration",
            }
        ]
        if erase_device_metadata:
            steps.append({
                "interface": "deploy",
                "step": "erase_devices_metadata"})
        self.manual_cleaning(node, clean_steps=steps)

    def rescue_unrescue(self):
        rescue_password = uuidutils.generate_uuid()
        self.rescue_node(self.node['uuid'], rescue_password)
        self.assertTrue(self.ping_ip_address(self.node_ip,
                                             should_succeed=True))

        # Open ssh connection to server
        linux_client = remote_client.RemoteClient(
            self.node_ip,
            'rescue',
            password=rescue_password,
            ssh_timeout=CONF.baremetal.rescue_timeout)
        linux_client.validate_authentication()

        self.unrescue_node(self.node['uuid'])
        self.assertTrue(self.ping_ip_address(self.node_ip,
                                             should_succeed=True))

    def boot_node_ramdisk(self, ramdisk_ref, iso=False):
        """Boot ironic using a ramdisk node.

        The following actions are executed:
          * Create/Pick networks to boot node in.
          * Create Neutron port and attach it to node.
          * Update node image_source.
          * Deploy node.
          * Wait until node is deployed.

        :param ramdisk_ref: Reference to user image or ramdisk to boot
                            the node with.
        :param iso: Boolean, default False, to indicate if the image ref
                    us actually an ISO image.
        """
        if ramdisk_ref is None:
            ramdisk_ref = self.image_ref

        network, subnet, router = self.create_networks()
        n_port = self.create_neutron_port(network_id=network['id'])
        self.vif_attach(node_id=self.node['uuid'], vif_id=n_port['id'])
        if iso:
            patch_path = '/instance_info/boot_iso'
        else:
            # NOTE(TheJulia): The non ISO ramdisk path supports this
            # and it being here makes it VERY easy for us to add a test
            # of just a kernel/ramdisk loading from glance at some point.
            patch_path = '/instance_info/image_source'
        patch = [{'path': patch_path,
                  'op': 'add',
                  'value': ramdisk_ref}]
        self.update_node(self.node['uuid'], patch=patch)
        self.set_node_provision_state(self.node['uuid'], 'active')
        self.addCleanup(
            self.set_node_provision_state,
            self.node['uuid'], 'deleted')
        if CONF.validation.connect_method == 'floating':
            self.node_ip = self.add_floatingip_to_node(self.node['uuid'])
        elif CONF.validation.connect_method == 'fixed':
            self.node_ip = self.get_server_ip(self.node['uuid'])
        else:
            m = ('Configuration option "[validation]/connect_method" '
                 'must be set.')
            raise lib_exc.InvalidConfiguration(m)
        self.wait_power_state(self.node['uuid'],
                              bm.BaremetalPowerStates.POWER_ON)
        self.wait_provisioning_state(self.node['uuid'],
                                     bm.BaremetalProvisionStates.ACTIVE,
                                     timeout=CONF.baremetal.active_timeout,
                                     interval=30)

    def boot_and_verify_ramdisk_node(self, ramdisk_ref=None, iso=False,
                                     should_succeed=True):
        node = self.get_node(node_id=self.node['uuid'])
        node_capabilities = node['properties'].get('capabilities', '')
        if (ramdisk_ref and 'tinycore' in ramdisk_ref.lower()
                and ('boot_mode:uefi' in node_capabilities
                     or 'uefi' in node.get('boot_mode', 'None'))):
            # Checks the top level field if the API gives it to us, i.e.
            # the test version supports it, and the lower level original
            # properties field which upstream CI populates.
            raise self.skipException('Skipping ramdisk test as tinycore '
                                     'cannot be booted in UEFI mode.')

        self.boot_node_ramdisk(ramdisk_ref, iso)
        self.assertTrue(self.ping_ip_address(self.node_ip,
                                             should_succeed=should_succeed))

    @classmethod
    def boot_node_anaconda(cls, image_ref, kernel_ref, ramdisk_ref,
                           stage2_ref=None):
        """Boot ironic using a ramdisk node.

        The following actions are executed:
          * Create/Pick networks to boot node in.
          * Create Neutron port and attach it to node.
          * Update node image_source.
          * Deploy node.
          * Wait until node is deployed.

        :param ramdisk_ref: Reference to user image or ramdisk to boot
                            the node with.
        :param iso: Boolean, default False, to indicate if the image ref
                    us actually an ISO image.
        """
        if image_ref is None or kernel_ref is None or ramdisk_ref is None:
            raise cls.skipException('Skipping anaconda tests as an image ref '
                                    'was not supplied')

        network, subnet, router = cls.create_networks()
        n_port = cls.create_neutron_port(network_id=network['id'])
        cls.vif_attach(node_id=cls.node['uuid'], vif_id=n_port['id'])
        p_root = '/instance_info/'
        patch = [{'path': p_root + 'image_source',
                  'op': 'add',
                  'value': image_ref},
                 {'path': p_root + 'kernel',
                  'op': 'add',
                  'value': kernel_ref},
                 {'path': p_root + 'ramdisk',
                  'op': 'add',
                  'value': ramdisk_ref}]
        if stage2_ref:
            patch.append(
                {
                    'path': p_root + 'stage2',
                    'op': 'add',
                    'value': stage2_ref,
                }
            )
        cls.update_node(cls.node['uuid'], patch=patch)
        cls.set_node_provision_state(cls.node['uuid'], 'active')
        if CONF.validation.connect_method == 'floating':
            cls.node_ip = cls.add_floatingip_to_node(cls.node['uuid'])
        elif CONF.validation.connect_method == 'fixed':
            cls.node_ip = cls.get_server_ip(cls.node['uuid'])
        else:
            m = ('Configuration option "[validation]/connect_method" '
                 'must be set.')
            raise lib_exc.InvalidConfiguration(m)
        cls.wait_power_state(cls.node['uuid'],
                             bm.BaremetalPowerStates.POWER_ON)

        if CONF.baremetal.anaconda_exit_test_at == 'heartbeat':
            cls.wait_for_agent_heartbeat(
                cls.node['uuid'],
                timeout=CONF.baremetal.anaconda_active_timeout)
        elif CONF.baremetal.anaconda_exit_test_at == 'active':
            cls.wait_provisioning_state(
                cls.node['uuid'],
                bm.BaremetalProvisionStates.ACTIVE,
                timeout=CONF.baremetal.anaconda_active_timeout,
                interval=30)

    def boot_and_verify_anaconda_node(self,
                                      image_ref=None,
                                      kernel_ref=None,
                                      ramdisk_ref=None,
                                      stage2_ref=None,
                                      should_succeed=True):
        self.boot_node_anaconda(image_ref, kernel_ref, ramdisk_ref)
        self.assertTrue(self.ping_ip_address(self.node_ip,
                                             should_succeed=should_succeed))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc

from ironic_tempest_plugin.tests.api.admin import api_microversion_fixture
from ironic_tempest_plugin.tests.api import base


class TestPortGroups(base.BaseBaremetalTest):
    """Basic positive test cases for port groups."""

    min_microversion = '1.26'  # portgroup mode and properties introduced

    def setUp(self):
        super(TestPortGroups, self).setUp()
        self.useFixture(
            api_microversion_fixture.APIMicroversionFixture(
                self.min_microversion))
        _, self.chassis = self.create_chassis()
        _, self.node = self.create_node(self.chassis['uuid'],
                                        network_interface='noop')
        _, self.portgroup = self.create_portgroup(
            self.node['uuid'], address=data_utils.rand_mac_address(),
            name=data_utils.rand_name('portgroup'))

    @decorators.idempotent_id('110cd302-256b-4ddc-be10-fc6c9ad8e649')
    def test_create_portgroup_with_address(self):
        """Create a port group with specific MAC address."""
        _, body = self.client.show_portgroup(self.portgroup['uuid'])
        self.assertEqual(self.portgroup['address'], body['address'])

    @decorators.idempotent_id('4336fa0f-da86-4cec-b788-89f59a7635a5')
    def test_create_portgroup_no_address(self):
        """Create a port group without setting MAC address."""
        _, portgroup = self.create_portgroup(self.node['uuid'])
        _, body = self.client.show_portgroup(portgroup['uuid'])

        self._assertExpected(portgroup, body)
        self.assertIsNone(body['address'])

    @decorators.idempotent_id('8378c69f-f806-454b-8ddd-6b7fd93ab12b')
    def test_delete_portgroup(self):
        """Delete a port group."""
        self.delete_portgroup(self.portgroup['uuid'])
        self.assertRaises(lib_exc.NotFound, self.client.show_portgroup,
                          self.portgroup['uuid'])

    @decorators.idempotent_id('f6be5e70-3e3b-435c-b2fc-bbb2cc9b3185')
    def test_show_portgroup(self):
        """Show a specified port group."""
        _, portgroup = self.client.show_portgroup(self.portgroup['uuid'])
        self._assertExpected(self.portgroup, portgroup)

    @decorators.idempotent_id('cf2dfd95-5ea1-4109-8ad3-297cd76aa5d3')
    def test_list_portgroups(self):
        """List port groups."""
        _, body = self.client.list_portgroups()
        self.assertIn(self.portgroup['uuid'],
                      [i['uuid'] for i in body['portgroups']])
        self.assertIn(self.portgroup['address'],
                      [i['address'] for i in body['portgroups']])
        self.assertIn(self.portgroup['name'],
                      [i['name'] for i in body['portgroups']])

    @decorators.idempotent_id('6a491006-2dd5-4c82-be39-a4fa015071c0')
    def test_update_portgroup_replace(self):
        """Update portgroup by replacing it's address and extra data."""
        new_address = data_utils.rand_mac_address()
        new_extra = {'foo': 'test'}

        patch = [{'path': '/address',
                  'op': 'replace',
                  'value': new_address},
                 {'path': '/extra/foo',
                  'op': 'replace',
                  'value': new_extra['foo']},
                 ]

        self.client.update_portgroup(self.portgroup['uuid'], patch)

        _, body = self.client.show_portgroup(self.portgroup['uuid'])

        self.assertEqual(new_address, body['address'])
        self._assertExpected(new_extra, body['extra'])

    @decorators.idempotent_id('9834a4ec-be41-40b4-a3a4-8e46ad7eb19d')
    def test_update_portgroup_remove_by_key(self):
        """Update portgroup by removing value from extra data."""
        self.client.update_portgroup(
            self.portgroup['uuid'],
            [{'path': '/extra/foo', 'op': 'remove'}]
        )
        _, body = self.client.show_portgroup(self.portgroup['uuid'])
        self.assertNotIn('foo', body['extra'])
        self._assertExpected({'open': 'stack'}, body['extra'])

    @decorators.idempotent_id('5da2f7c7-03e8-4db0-8c3e-2fe6ebcc4439')
    def test_update_portgroup_remove_collection(self):
        """Update portgroup by removing collection from extra data."""
        self.client.update_portgroup(
            self.portgroup['uuid'],
            [{'path': '/extra', 'op': 'remove'}]
        )
        _, body = self.client.show_portgroup(self.portgroup['uuid'])
        self.assertEqual({}, body['extra'])

    @decorators.idempotent_id('a1123416-7bb6-4a6a-9f14-859c72550552')
    def test_update_portgroup_add(self):
        """Update portgroup by adding new extra data and properties."""
        patch = [{'path': '/extra/key1',
                  'op': 'add',
                  'value': 'value1'},
                 {'path': '/properties/key1',
                  'op': 'add',
                  'value': 'value1'}]

        self.client.update_portgroup(self.portgroup['uuid'], patch)

        _, body = self.client.show_portgroup(self.portgroup['uuid'])
        self._assertExpected({'key1': 'value1'}, body['extra'])
        self._assertExpected({'key1': 'value1'}, body['properties'])

    @decorators.idempotent_id('67d5013e-5158-44e8-8d1c-a01d04542be4')
    def test_update_portgroup_mixed_ops(self):
        """Update port group with add, replace and remove ops in one patch."""
        new_address = data_utils.rand_mac_address()
        new_extra = {'key3': {'cat': 'meow'}}

        patch = [{'path': '/address',
                  'op': 'replace',
                  'value': new_address},
                 {'path': '/extra/foo',
                  'op': 'remove'},
                 {'path': '/extra/key3',
                  'op': 'add',
                  'value': new_extra['key3']}]

        self.client.update_portgroup(self.portgroup['uuid'], patch)

        _, body = self.client.show_portgroup(self.portgroup['uuid'])
        self.assertEqual(new_address, body['address'])
        self._assertExpected(new_extra, body['extra'])
        self.assertNotIn('foo', body['extra'])


class TestPortGroupsWithJsonExtSupport(base.BaseBaremetalTest):
    """Basic test cases to validate appending .json extension to a portgroup's

    `name` or `uuid` in API versions 1.90 and prior works.
    """

    max_microversion = '1.90'  # Max API version allowed for testing: 1.90
    min_microversion = '1.90'  # Min API version allowed for testing: 1.90

    def setUp(self):
        super(TestPortGroupsWithJsonExtSupport, self).setUp()
        self.useFixture(
            api_microversion_fixture.APIMicroversionFixture(
                self.min_microversion))
        _, self.chassis = self.create_chassis()
        _, self.node = self.create_node(self.chassis['uuid'])
        _, self.portgroup = self.create_portgroup(
            self.node['uuid'], address=data_utils.rand_mac_address(),
            name=data_utils.rand_name('portgroup'))

    @decorators.idempotent_id('607ea131-ef5c-44e9-a900-a8e426b804e8')
    def test_delete_portgroup_with_json(self):
        """Delete a portgroup while appending .json extension to its uuid"""
        self.delete_portgroup('%s.json' % self.portgroup['uuid'])
        self.assertRaises(lib_exc.NotFound, self.client.show_portgroup,
                          self.portgroup['uuid'])

    @decorators.idempotent_id('9126ecac-310f-440a-81d4-ba9af1767845')
    def test_show_portgroup_with_json(self):
        """Show a portgroup while appending .json extension to its uuid"""
        _, portgroup = self.client.show_portgroup('%s.json' %
                                                  self.portgroup['uuid'])
        self._assertExpected(self.portgroup, portgroup)

    @decorators.idempotent_id('a6f154aa-e447-4576-89e2-9e8951fb7c43')
    def test_update_portgroup_with_json(self):
        """Update a portgroup while appending .json extension to its uuid"""
        new_address = data_utils.rand_mac_address()
        new_extra = {'foo': 'test'}

        patch = [{'path': '/address',
                  'op': 'replace',
                  'value': new_address},
                 {'path': '/extra/foo',
                  'op': 'replace',
                  'value': new_extra['foo']},
                 ]

        self.client.update_portgroup('%s.json' % self.portgroup['uuid'],
                                     patch)

        _, body = self.client.show_portgroup(self.portgroup['uuid'])

        self.assertEqual(new_address, body['address'])
        self._assertExpected(new_extra, body['extra'])


class TestPortGroupsWithoutJsonExtSupport(base.BaseBaremetalTest):
    """Basic test cases to validate appending .json extension to a portgroup's

    `name` or `uuid` in API versions later than 1.90 is disabled.
    """

    min_microversion = '1.91'  # Min API version allowed for testing: 1.91

    def setUp(self):
        super(TestPortGroupsWithoutJsonExtSupport, self).setUp()
        _, self.chassis = self.create_chassis()
        _, self.node = self.create_node(self.chassis['uuid'])
        _, self.portgroup = self.create_portgroup(
            self.node['uuid'], address=data_utils.rand_mac_address(),
            name=data_utils.rand_name('portgroup'))

    @decorators.idempotent_id('2aede448-7165-4d54-92de-3be31ae19af0')
    def test_delete_portgroup_with_json(self):
        """Trying to delete a portgroup while appending .json ext 404s"""
        self.assertRaises(lib_exc.NotFound, self.client.delete_portgroup,
                          '%s.json' % self.portgroup['uuid'])

    @decorators.idempotent_id('be8ade56-8ac1-49f4-981a-93b403fa1d79')
    def test_show_portgroup_with_json(self):
        """Trying to show a portgroup while appending .json ext 404s"""
        self.assertRaises(lib_exc.NotFound, self.client.show_portgroup,
                          '%s.json' % self.portgroup['uuid'])

    @decorators.idempotent_id('f624352a-6560-4ec1-b4ff-ccfe7cda08f2')
    def test_update_portgroup_with_json(self):
        """Trying to update a portgroup while appending .json ext 404s"""
        new_address = data_utils.rand_mac_address()
        new_extra = {'foo': 'test'}

        patch = [{'path': '/address',
                  'op': 'replace',
                  'value': new_address},
                 {'path': '/extra/foo',
                  'op': 'replace',
                  'value': new_extra['foo']},
                 ]

        self.assertRaises(lib_exc.NotFound, self.client.update_portgroup,
                          '%s.json' % self.portgroup['uuid'], patch)



#
# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging
from tempest.common import utils
from tempest import config
from tempest.lib.common.utils import data_utils
from tempest.lib.common.utils import test_utils
from tempest.lib import decorators

from ironic_tempest_plugin import manager
from ironic_tempest_plugin.tests.scenario import baremetal_manager

LOG = logging.getLogger(__name__)
CONF = config.CONF


class BaremetalMultitenancy(baremetal_manager.BaremetalScenarioTest,
                            manager.NetworkScenarioTest):
    """Check L2 isolation of baremetal and VM instances in different tenants:

    * Create a keypair, network, subnet and router for the primary tenant
    * Boot 2 instances in the different tenant's network using the keypair
    * Associate floating ips to both instance
    * Verify there is no L3 connectivity between instances of different tenants
    * Verify connectivity between instances floating IP's
    * Delete both instances
    """

    credentials = ['primary', 'alt', 'admin', 'system_admin']

    @classmethod
    def skip_checks(cls):
        super(BaremetalMultitenancy, cls).skip_checks()
        if not CONF.baremetal.use_provision_network:
            msg = 'Ironic/Neutron tenant isolation is not configured.'
            raise cls.skipException(msg)

    def create_tenant_network(self, clients, tenant_cidr, create_router=True):
        network = self.create_network(
            networks_client=clients.networks_client,
            project_id=clients.credentials.project_id)

        router = None
        if create_router:
            router = self.get_router(
                client=clients.routers_client,
                project_id=clients.credentials.tenant_id)

        result = clients.subnets_client.create_subnet(
            name=data_utils.rand_name('subnet'),
            network_id=network['id'],
            tenant_id=clients.credentials.tenant_id,
            ip_version=4,
            cidr=tenant_cidr)
        subnet = result['subnet']
        if create_router:
            clients.routers_client.add_router_interface(router['id'],
                                                        subnet_id=subnet['id'])
        self.addCleanup(clients.subnets_client.delete_subnet, subnet['id'])

        if create_router:
            self.addCleanup(clients.routers_client.remove_router_interface,
                            router['id'], subnet_id=subnet['id'])
        return network, subnet, router

    def verify_l3_connectivity(self, source_ip, private_key,
                               destination_ip, conn_expected=True, timeout=15):
        remote = self.get_remote_client(source_ip, private_key=private_key)
        remote.validate_authentication()

        output = remote.exec_command('ip route')
        LOG.debug("Routing table on %s is %s", source_ip, output)

        cmd = 'ping %s -c4 -w4 || exit 0' % destination_ip
        success_substring = " bytes from %s" % destination_ip

        def ping_remote():
            output = remote.exec_command(cmd)
            LOG.debug("Got output %s while pinging %s", output, destination_ip)
            if conn_expected:
                return success_substring in output
            else:
                return success_substring not in output

        # NOTE(vsaienko): we may lost couple of pings due to missing ARPs
        # so do several retries to get stable output.
        res = test_utils.call_until_true(ping_remote, timeout, 1)
        self.assertTrue(res)

    def multitenancy_check(self, use_vm=False):
        tenant_cidr = '10.0.100.0/24'
        keypair = self.create_keypair()
        network, subnet, router = self.create_tenant_network(
            self.os_primary, tenant_cidr)
        alt_keypair = self.create_keypair(self.os_alt.keypairs_client)
        alt_network, alt_subnet, alt_router = self.create_tenant_network(
            self.os_alt, tenant_cidr)
        # Create single BM guest as Primary
        instance1, node1 = self.boot_instance(
            clients=self.os_primary,
            keypair=keypair,
            net_id=network['id'],
            fixed_ip='10.0.100.101',
        )
        fixed_ip1 = instance1['addresses'][network['name']][0]['addr']
        floating_ip1 = self.create_floating_ip(
            instance1,
        )['floating_ip_address']
        self.check_vm_connectivity(ip_address=floating_ip1,
                                   private_key=keypair['private_key'],
                                   server=instance1)
        if use_vm:
            # Create VM on compute node
            alt_instance = self.create_server(
                clients=self.os_alt,
                key_name=alt_keypair['name'],
                flavor=CONF.compute.flavor_ref_alt,
                networks=[{'uuid': alt_network['id'],
                           'fixed_ip': '10.0.100.102'}],
            )
        else:
            # Create BM
            alt_instance, alt_node = self.boot_instance(
                keypair=alt_keypair,
                clients=self.os_alt,
                net_id=alt_network['id'],
                fixed_ip='10.0.100.102',
            )
        fixed_ip2 = alt_instance['addresses'][alt_network['name']][0]['addr']
        alt_floating_ip = self.create_floating_ip(
            alt_instance,
            client=self.os_alt.floating_ips_client
        )['floating_ip_address']
        self.check_vm_connectivity(
            ip_address=alt_floating_ip,
            private_key=alt_keypair['private_key'],
            server=alt_instance)
        self.verify_l3_connectivity(
            alt_floating_ip,
            alt_keypair['private_key'],
            fixed_ip1,
            conn_expected=False
        )
        self.verify_l3_connectivity(
            floating_ip1,
            keypair['private_key'],
            fixed_ip2,
            conn_expected=False
        )
        self.verify_l3_connectivity(
            floating_ip1,
            keypair['private_key'],
            alt_floating_ip,
            conn_expected=True
        )
        self.terminate_instance(
            instance=alt_instance,
            servers_client=self.os_alt.servers_client)
        self.terminate_instance(instance=instance1)

    @decorators.idempotent_id('26e2f145-2a8e-4dc7-8457-7f2eb2c6749d')
    @utils.services('compute', 'image', 'network')
    def test_baremetal_multitenancy(self):
        self.multitenancy_check()

    @decorators.idempotent_id('9e38631a-2df2-11e9-810e-8c16450ea513')
    @utils.services('compute', 'image', 'network')
    def test_baremetal_vm_multitenancy(self):
        if not CONF.service_available.nova:
            self.skipTest('Compute service Nova is disabled,'
                          ' VM is required to run this test')
        self.multitenancy_check(use_vm=True)

    @decorators.idempotent_id('6891929f-a254-43b1-bd97-6ea3ec74d6a9')
    @utils.services('compute', 'image', 'network')
    def test_baremetal_vm_multitenancy_trunk(self):
        """Check Trunk scenario for two baremetal servers


        fipA -- RouterA -- NetworkA (10.0.100.0/24)
                              |
                              |eth0
                             eth0:instanceA
                              |eth0.vlan_id
                              |
                           NetworkB(10.0.101.0/24)
                              |
                              |eth0
                           instanceB


        * Create instanceA within networkA and FIPA with trunk port plugged to
          networkA as parent(native vlan/untagged) and networkB as
          vlan subport
        * Create instanceB within networkB
        * Verify connectivity to instanceB from instanceA failed
        * Assign ip address on subport inside instanceA. This step is needed
          only unless nova configdrive support of trunks is not implemented.
        * Verify connectivity to instanceB from instanceA success
        * Remove subport from instanceA
        * Verify connectivity to instanceB from instanceA failed
        * Add subport to instanceA
        * Verify connectivity to instanceB from instanceA Passed
        """

        if not CONF.baremetal_feature_enabled.trunks_supported:
            msg = 'Trunks with baremetal are not supported.'
            raise self.skipException(msg)

        tenant_a_cidr = '10.0.100.0/24'
        tenant_b_cidr = '10.0.101.0/24'

        keypair = self.create_keypair()
        networkA, subnetA, routerA = self.create_tenant_network(
            self.os_primary, tenant_a_cidr)
        networkB, subnetB, _ = self.create_tenant_network(
            self.os_primary, tenant_b_cidr, create_router=False)
        portB = self.create_port(network_id=networkB["id"])

        parent_port = self.create_port(network_id=networkA["id"])
        subport = self.create_port(network_id=networkB["id"])
        subports = [{'port_id': subport['id'], 'segmentation_type': 'inherit'}]
        trunk = self.os_primary.trunks_client.create_trunk(
            name="test-trunk", port_id=parent_port['id'],
            sub_ports=subports)['trunk']
        self.addCleanup(self.os_primary.trunks_client.delete_trunk,
                        trunk['id'])

        # Create instanceB first as we will not check if its booted,
        # as we don't have FIP, so it has more time than instanceA
        # to boot.
        instanceB, nodeB = self.boot_instance(
            clients=self.os_primary,
            keypair=keypair,
            networks=[{'port': portB['id']}]
        )

        instanceA, nodeA = self.boot_instance(
            clients=self.os_primary,
            keypair=keypair,
            networks=[{'port': parent_port['id']}]
        )

        floating_ipA = self.create_floating_ip(
            instanceA,
        )['floating_ip_address']

        fixed_ipB = instanceB['addresses'][networkB['name']][0]['addr']

        self.check_vm_connectivity(ip_address=floating_ipA,
                                   private_key=keypair['private_key'],
                                   server=instanceA)
        ssh_client = self.get_remote_client(floating_ipA,
                                            private_key=keypair['private_key'])

        # TODO(vsaienko): add when cloudinit support is implemented
        # add validation of network_data.json and drop next ip assignment

        self.verify_l3_connectivity(
            floating_ipA,
            keypair['private_key'],
            fixed_ipB,
            conn_expected=False
        )
        vlan_id = trunk['sub_ports'][0]['segmentation_id']
        subport_ip = subport['fixed_ips'][0]['ip_address']

        interface_name = ssh_client.exec_command(
            "sudo ip route | awk '/default/ {print $5}'").rstrip()
        cmds = [
            f"sudo ip link add link {interface_name} name "
            f"{interface_name}.{vlan_id} type vlan id {vlan_id}",
            f"sudo ip addr add {subport_ip}/24 dev {interface_name}.{vlan_id}",
            f"sudo ip link set dev {interface_name}.{vlan_id} up"]

        for cmd in cmds:
            ssh_client.exec_command(cmd)

        self.verify_l3_connectivity(
            floating_ipA,
            keypair['private_key'],
            fixed_ipB,
            conn_expected=True
        )

        self.os_primary.trunks_client.delete_subports_from_trunk(
            trunk['id'], trunk['sub_ports'])
        self.verify_l3_connectivity(
            floating_ipA,
            keypair['private_key'],
            fixed_ipB,
            conn_expected=False
        )
        self.os_primary.trunks_client.add_subports_to_trunk(
            trunk['id'], trunk['sub_ports'])

        # NOTE(vsaienko): it may take some time for network driver to
        # setup vlans as this is async operation.
        self.verify_l3_connectivity(
            floating_ipA,
            keypair['private_key'],
            fixed_ipB,
            conn_expected=True
        )
        self.terminate_instance(instance=instanceA)
        self.terminate_instance(instance=instanceB)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest import config
# from tempest.lib.common import rest_client
# from tempest.lib.common.utils import data_utils
# from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc

from ironic_tempest_plugin.common import waiters
from ironic_tempest_plugin.tests.api import base

CONF = config.CONF


class TestNodeProjectReader(base.BaseBaremetalRBACTest):
    """Tests for baremetal nodes with a tempest project reader."""

    credentials = ['system_admin', 'project_reader']

    def setUp(self):
        super(TestNodeProjectReader, self).setUp()

        self.client = self.os_system_admin.baremetal.BaremetalClient()
        self.reader_client = self.os_project_reader.baremetal.BaremetalClient()
        _, self.chassis = self.create_chassis()
        # Bare node, no inherent permissions by default for project readers.
        _, self.node = self.create_node(self.chassis['uuid'],
                                        network_interface='noop',
                                        deploy_interface='fake')

    # Default policy is:
    # ('role:reader and '
    #  '(project_id:%(node.owner)s or project_id:%(node.lessee)s)')

    def test_reader_cannot_create_node(self):
        try:
            resp, body = self.reader_client.create_node(self.chassis['uuid'])
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_get_node(self):
        """Reader cannot get node

        baremetal:node:list
        """
        try:
            resp, body = self.reader_client.show_node(self.node['uuid'])
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_list_is_empty(self):
        """Expected default for no lessee or owner rights is an empty list.

        baremetal:node:list and baremetal:node:list_all
        """
        resp, body = self.reader_client.list_nodes()
        self.assertEqual(0, len(body['nodes']))

    def test_reader_cannot_update_node(self):
        """Reader cannot update node

        baremetal:node:update
        """
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'])
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_update_node_driver_info(self):
        """Reader cannot update node driver_info

        baremetal:node:update:driver_info
        """
        patch = [{'path': '/driver_info/ipmi_username', 'op': 'replace',
                  'value': 'foo_user'}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_update_node_properties(self):
        """Reader cannot update node properties

        baremetal:node:update:properties
        """
        new_p = {'cpu_arch': 'arm64', 'cpus': '1', 'local_gb': '10000',
                 'memory_mb': '12300'}
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        properties=new_p)
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_update_node_chassis_uuid(self):
        """Reader cannot update node chassis uuid

        baremetal:node:update:chassis_uuid
        """
        patch = [{'path': '/chassis_uuid', 'op': 'replace',
                  'value': 'new_chassis_uuid'}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_update_node_instance_uuid(self):
        """Reader cannot update node instance uuid

        baremetal:node:update:instance_uuid
        """
        patch = [{'path': '/instance_uuid', 'op': 'replace',
                  'value': 'new_instance_uuid'}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_update_node_lessee(self):
        """Reader cannot update node lessee

        baremetal:node:update:lessee
        """
        patch = [{'path': '/lessee', 'op': 'replace',
                  'value': 'new_lessee'}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_update_node_owner(self):
        """Reader cannot update node owner

        baremetal:node:update:owner
        """
        patch = [{'path': '/owner', 'op': 'replace',
                  'value': 'new_owner'}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_update_node_driver_interfaces(self):
        """Reader cannot update node driver interfaces

        baremetal:node:update:driver_interfaces
        """
        patch = [{'path': '/driver', 'op': 'replace', 'value': 'ipmi'}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_update_node_network_data(self):
        """Reader cannot update node network data

        baremetal:node:update:network_data
        """
        new_net_data = {'networks': [], 'services': [], 'links': []}
        patch = [{'path': '/network_data', 'op': 'replace',
                  'value': new_net_data}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_update_node_conductor_group(self):
        """Reader cannot update node conductor group

        baremetal:node:update:conductor_group
        """
        patch = [{'path': '/conductor_group', 'op': 'replace',
                  'value': 'new_group'}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_update_node_name(self):
        """Reader cannot update node name

        baremetal:node:update:name
        """
        patch = [{'path': '/name', 'op': 'replace', 'value': 'new_name'}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_update_node_retired(self):
        """Reader cannot update node retired

        baremetal:node:update:retired
        """
        patch = [{'path': '/retired', 'op': 'replace', 'value': True}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_update_node_extra(self):
        """Reader cannot update node extra

        baremetal:node:update:extra
        """
        patch = [{'path': '/extra', 'op': 'replace',
                  'value': {'extra': 'extra'}}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_update_instance_info(self):
        """Reader cannot update node instance info

        baremetal:node:update:instance_info
        """
        patch = [{'path': '/instance_info', 'op': 'replace',
                  'value': {'display_name': 'new_display_name'}}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_update_owner_provisioned(self):
        """Reader cannot update node owner provisioned

        baremetal:node:update_owner_provisioned
        """
        provision_states_list = ['manage', 'provide', 'active']
        target_states_list = ['manageable', 'available', 'active']
        for (provision_state, target_state) in zip(provision_states_list,
                                                   target_states_list):
            self.client.set_node_provision_state(self.node['uuid'],
                                                 provision_state)
            waiters.wait_for_bm_node_status(self.client, self.node['uuid'],
                                            attr='provision_state',
                                            status=target_state, timeout=10)

        patch = [{'path': '/owner', 'op': 'replace',
                  'value': {'display_name': 'new_owner'}}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.NotFound as e:
            resp = e.resp
        finally:
            self.client.set_node_provision_state(self.node['uuid'],
                                                 'deleted')

        self.assertEqual(404, resp.status)

    def test_reader_cannot_delete_node(self):
        """Reader cannot delete node

        baremetal:node:delete
        """
        try:
            resp, body = self.reader_client.delete_node(self.node['uuid'])
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_validate_node(self):
        """Reader cannot validate node

        baremetal:node:validate
        """
        try:
            resp, body = self.reader_client.validate_driver_interface(
                self.node['uuid'])
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_set_maintenance(self):
        """Reader cannot set maintenance mode

        baremetal:node:set_maintenance
        """
        patch = [{'path': '/maintenance', 'op': 'replace', 'value': True}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_unset_maintenance(self):
        """Reader cannot unset maintenance mode

        baremetal:node:clear_maintenance
        """
        patch = [{'path': '/maintenance', 'op': 'replace', 'value': False}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_get_boot_device(self):
        """Reader cannot get boot device

        baremetal:node:get_boot_device
        """
        try:
            resp, body = self.reader_client.get_node_boot_device(
                self.node['uuid'])
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_set_boot_device(self):
        """Reader cannot set boot device

        baremetal:node:set_boot_device
        """
        try:
            resp, body = self.reader_client.set_node_boot_device(
                self.node['uuid'], 'pxe')
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_get_indicator_state(self):
        """Reader cannot get indicator state

        baremetal:node:get_indicator_state
        """
        try:
            resp, body = self.reader_client.get_node_indicator_state(
                self.node['uuid'], 'system', 'led')
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_set_indicator_state(self):
        """Reader cannot set indicator state

        baremetal:node:set_indicator_state
        """
        try:
            resp, body = self.reader_client.set_node_indicator_state(
                self.node['uuid'], 'system', 'led', 'BLINKING')
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_inject_nmi(self):
        """Reader cannot inject NMI

        baremetal:node:inject_nmi
        """
        try:
            resp, body = self.reader_client._put_request(
                '/v1/nodes/%s/management/inject_nmi' % self.node['uuid'], {})
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_get_states(self):
        """Reader cannot list the states of the node

        baremetal:node:get_states
        """
        try:
            resp, body = self.reader_client.list_nodestates(self.node['uuid'])
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_set_power_state(self):
        """Reader cannot set power state

        baremetal:node:set_power_state
        """
        try:
            resp, body = self.reader_client.set_node_power_state(
                self.node['uuid'], 'off')
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_set_boot_mode(self):
        """Reader cannot set boot mode

        baremetal:node:set_boot_mode
        """
        try:
            resp, body = self.reader_client.set_node_state(
                self.node['uuid'], 'boot_mode', 'uefi')
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_set_secure_boot(self):
        """Reader cannot set secure boot

        baremetal:node:set_secure_boot
        """
        try:
            resp, body = self.reader_client.set_node_state(
                self.node['uuid'], 'secure_boot', True)
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_set_provision_state(self):
        """Reader cannot set provision state

        baremetal:node:set_provision_state
        """
        try:
            resp, body = self.reader_client.set_node_provision_state(
                self.node['uuid'], 'manage')
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_set_raid_state(self):
        """Reader cannot set raid state

        baremetal:node:set_raid_state
        """
        try:
            resp, body = self.reader_client.set_node_state(
                self.node['uuid'], 'target_raid_config', {'raid': 'config'})
        except lib_exc.UnexpectedResponseCode as e:
            resp = e.resp

        self.assertEqual(405, resp.status)

    def test_reader_cannot_get_console(self):
        """Reader cannot get console

        baremetal:node:get_console
        """
        try:
            resp, body = self.reader_client.get_console(self.node['uuid'])
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_set_console_state(self):
        """Reader cannot set console state

        baremetal:node:set_console_state
        """
        try:
            resp, body = self.reader_client.set_console_mode(
                self.node['uuid'], enabled=True)
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_vif_list(self):
        """Reader cannot list vifs

        baremetal:node:vif:list
        """
        try:
            resp, body = self.reader_client.vif_list(self.node['uuid'])
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_vif_attach(self):
        """Reader cannot attach vif

        baremetal:node:vif:attach
        """
        try:
            resp = self.reader_client.vif_attach(
                self.node['uuid'], 'vifid')
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_vif_detach(self):
        """Reader cannot detach vif

        baremetal:node:vif:detach
        """
        try:
            resp, body = self.reader_client.vif_detach(
                self.node['uuid'], 'vifid')
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_traits_list(self):
        """Reader cannot list traits

        baremetal:node:traits:list
        """
        try:
            resp, body = self.reader_client.list_node_traits(self.node['uuid'])
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_traits_set(self):
        """Reader cannot set traits

        baremetal:node:traits:set
        """
        try:
            resp, body = self.reader_client.set_node_traits(
                self.node['uuid'], ['CUSTOM_TRAIT_A', 'CUSTOM_TRAIT_B'])
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_traits_delete(self):
        """Reader cannot delete traits

        baremetal:node:traits:delete
        """
        try:
            resp, body = self.reader_client.remove_node_traits(
                self.node['uuid'])
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_bios_get(self):
        """Reader cannot get bios settings

        baremetal:node:bios:get
        """
        try:
            resp, body = self.reader_client.list_node_bios_settings(
                self.node['uuid'])
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_disable_cleaning(self):
        """Reader cannot disable automated node cleaning

        baremetal:node:disable_cleaning
        """
        patch = [{'path': '/automated_clean', 'op': 'replace', 'value': False}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_history_get(self):
        """Reader cannot list history entries for a node

        baremetal:node:history:get
        """
        try:
            resp, body = self.reader_client.list_node_history(
                self.node['uuid'])
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_vendor_passthru(self):
        """Reader cannot list vendor-specific extensions

        baremetal:node:vendor_passthru
        """
        try:
            resp, body = self.reader_client.list_vendor_passthru_methods(
                self.node['uuid'])
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_ipa_heartbeat(self):
        """Reader cannot heartbeat

        baremetal:node:ipa_heartbeat
        """
        # TODO(hjensas)
        # try:
        #     resp, body = self.reader_client.ipa_heartbeat(
        #         self.node['uuid'], callback_url='http://foo/',
        #         agent_token=uuidutils.generate_uuid(), agent_version='1')
        # except lib_exc.BadRequest as e:
        #     resp = e.resp
        #
        # self.assertEqual(400, resp.status)
        pass


class TestNodeSystemReader(base.BaseBaremetalRBACTest):
    """Tests for baremetal nodes with a tempest system reader.

    All tests here must always expect *multiple* nodes visible, since
    this is a global reader role.

    https://opendev.org/openstack/ironic/src/branch/master/ironic/common/policy.py#L60
    """

    credentials = ['system_admin', 'system_reader']

    def setUp(self):
        super(TestNodeSystemReader, self).setUp()

        self.client = self.os_system_admin.baremetal.BaremetalClient()
        self.reader_client = self.os_system_reader.baremetal.BaremetalClient()
        _, self.chassis = self.create_chassis()
        _, self.node = self.create_node(self.chassis['uuid'],
                                        network_interface='noop',
                                        deploy_interface='fake')

    def test_reader_cannot_create_node(self):
        """Reader cannot create node

        baremetal:node:create
        """
        try:
            resp, body = self.reader_client.create_node(self.chassis['uuid'])
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_can_get_node(self):
        """Reader can get node

        baremetal:node:??get??list??
        """
        resp, body = self.reader_client.show_node(self.node['uuid'])
        self.assertEqual(200, resp.status)

    def test_reader_list_is_not_empty(self):
        """List nodes return all nodes

        baremetal:node:list and baremetal:node:list_all
        """
        resp, body = self.reader_client.list_nodes()
        self.assertGreater(len(body['nodes']), 0)

    def test_reader_cannot_update_node(self):
        """Reader cannot update node

        baremetal:node:update
        """
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'])
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_update_node_driver_info(self):
        """Reader cannot update node driver_info

        baremetal:node:update:driver_info
        """
        patch = [{'path': '/driver_info/ipmi_username', 'op': 'replace',
                  'value': 'foo_user'}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_update_node_properties(self):
        """Reader cannot update node properties

        baremetal:node:update:properties
        """
        new_p = {'cpu_arch': 'arm64', 'cpus': '1', 'local_gb': '10000',
                 'memory_mb': '12300'}
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        properties=new_p)
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_update_node_chassis_uuid(self):
        """Reader cannot update node chassis uuid

        baremetal:node:update:chassis_uuid
        """
        patch = [{'path': '/chassis_uuid', 'op': 'replace',
                  'value': 'new_chassis_uuid'}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_update_node_instance_uuid(self):
        """Reader cannot update node instance uuid

        baremetal:node:update:instance_uuid
        """
        patch = [{'path': '/instance_uuid', 'op': 'replace',
                  'value': 'new_instance_uuid'}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_update_node_lessee(self):
        """Reader cannot update node lessee

        baremetal:node:update:lessee
        """
        patch = [{'path': '/lessee', 'op': 'replace',
                  'value': 'new_lessee'}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_update_node_owner(self):
        """Reader cannot update node owner

        baremetal:node:update:owner
        """
        patch = [{'path': '/owner', 'op': 'replace',
                  'value': 'new_owner'}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_update_node_driver_interfaces(self):
        """Reader cannot update node driver interfaces

        baremetal:node:update:driver_interfaces
        """
        patch = [{'path': '/driver', 'op': 'replace', 'value': 'ipmi'}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_update_node_network_data(self):
        """Reader cannot update node network data

        baremetal:node:update:network_data
        """
        new_net_data = {'networks': [], 'services': [], 'links': []}
        patch = [{'path': '/network_data', 'op': 'replace',
                  'value': new_net_data}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_update_node_conductor_group(self):
        """Reader cannot update node conductor group

        baremetal:node:update:conductor_group
        """
        patch = [{'path': '/conductor_group', 'op': 'replace',
                  'value': 'new_group'}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_update_node_name(self):
        """Reader cannot update node name

        baremetal:node:update:name
        """
        patch = [{'path': '/name', 'op': 'replace', 'value': 'new_name'}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_update_node_retired(self):
        """Reader cannot update node retired

        baremetal:node:update:retired
        """
        patch = [{'path': '/retired', 'op': 'replace', 'value': True}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_update_node_extra(self):
        """Reader cannot update node extra

        baremetal:node:update:extra
        """
        patch = [{'path': '/extra', 'op': 'replace',
                  'value': {'extra': 'extra'}}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_update_instance_info(self):
        """Reader cannot update node instance info

        baremetal:node:update:instance_info
        """
        patch = [{'path': '/instance_info', 'op': 'replace',
                  'value': {'display_name': 'new_display_name'}}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_update_owner_provisioned(self):
        """Reader cannot update node owner provisioned

        baremetal:node:update_owner_provisioned
        """
        provision_states_list = ['manage', 'provide', 'active']
        target_states_list = ['manageable', 'available', 'active']
        for (provision_state, target_state) in zip(provision_states_list,
                                                   target_states_list):
            self.client.set_node_provision_state(self.node['uuid'],
                                                 provision_state)
            waiters.wait_for_bm_node_status(self.client, self.node['uuid'],
                                            attr='provision_state',
                                            status=target_state, timeout=10)

        patch = [{'path': '/owner', 'op': 'replace',
                  'value': {'display_name': 'new_owner'}}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.Forbidden as e:
            resp = e.resp
        finally:
            self.client.set_node_provision_state(self.node['uuid'],
                                                 'deleted')

        self.assertEqual(403, resp.status)

    def test_reader_cannot_delete_node(self):
        """Reader cannot delete node

        baremetal:node:delete
        """
        try:
            resp, body = self.reader_client.delete_node(self.node['uuid'])
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_validate_node(self):
        """Reader cannot validate node

        baremetal:node:validate
        """
        try:
            resp, body = self.reader_client.validate_driver_interface(
                self.node['uuid'])
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_set_maintenance(self):
        """Reader cannot set maintenance mode

        baremetal:node:set_maintenance
        """
        patch = [{'path': '/maintenance', 'op': 'replace', 'value': True}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_unset_maintenance(self):
        """Reader cannot unset maintenance mode

        baremetal:node:clear_maintenance
        """
        patch = [{'path': '/maintenance', 'op': 'replace', 'value': False}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_get_boot_device(self):
        """Reader cannot get boot device

        baremetal:node:get_boot_device
        """
        try:
            resp, body = self.reader_client.get_node_boot_device(
                self.node['uuid'])
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_set_boot_device(self):
        """Reader cannot set boot device

        baremetal:node:set_boot_device
        """
        try:
            resp, body = self.reader_client.set_node_boot_device(
                self.node['uuid'], 'pxe')
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_can_get_indicator_state(self):
        """Reader can get indicator state

        baremetal:node:get_indicator_state
        """
        resp, body = self.reader_client.get_node_indicator_state(
            self.node['uuid'], 'system', 'led')
        self.assertEqual(200, resp.status)

    def test_reader_cannot_set_indicator_state(self):
        """Reader cannot set indicator state

        baremetal:node:set_indicator_state
        """
        try:
            resp, body = self.reader_client.set_node_indicator_state(
                self.node['uuid'], 'system', 'led', 'BLINKING')
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_inject_nmi(self):
        """Reader cannot inject NMI

        baremetal:node:inject_nmi
        """
        try:
            resp, body = self.reader_client._put_request(
                '/v1/nodes/%s/management/inject_nmi' % self.node['uuid'], {})
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_can_get_states(self):
        """Reader can list the states of the node

        baremetal:node:get_states
        """
        resp, body = self.reader_client.list_nodestates(self.node['uuid'])
        self.assertEqual(200, resp.status)

    def test_reader_cannot_set_power_state(self):
        """Reader cannot set power state

        baremetal:node:set_power_state
        """
        try:
            resp, body = self.reader_client.set_node_power_state(
                self.node['uuid'], 'off')
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_set_boot_mode(self):
        """Reader cannot set boot mode

        baremetal:node:set_boot_mode
        """
        try:
            resp, body = self.reader_client.set_node_state(
                self.node['uuid'], 'boot_mode', 'uefi')
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_set_secure_boot(self):
        """Reader cannot set secure boot

        baremetal:node:set_secure_boot
        """
        try:
            resp, body = self.reader_client.set_node_state(
                self.node['uuid'], 'secure_boot', True)
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_set_provision_state(self):
        """Reader cannot set provision state

        baremetal:node:set_provision_state
        """
        try:
            resp, body = self.reader_client.set_node_provision_state(
                self.node['uuid'], 'manage')
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_set_raid_state(self):
        """Reader cannot set raid state

        baremetal:node:set_raid_state
        """
        try:
            resp, body = self.reader_client.set_node_state(
                self.node['uuid'], 'target_raid_config', {'raid': 'config'})
        except lib_exc.UnexpectedResponseCode as e:
            resp = e.resp

        self.assertEqual(405, resp.status)

    def test_reader_cannot_get_console(self):
        """Reader cannot get console

        baremetal:node:get_console
        """
        try:
            resp, body = self.reader_client.get_console(self.node['uuid'])
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_set_console_state(self):
        """Reader cannot set console state

        baremetal:node:set_console_state
        """
        try:
            resp, body = self.reader_client.set_console_mode(
                self.node['uuid'], enabled=True)
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_can_vif_list(self):
        """Reader can list vifs

        baremetal:node:vif:list
        """
        resp, body = self.reader_client.vif_list(self.node['uuid'])
        self.assertEqual(200, resp.status)

    def test_reader_cannot_vif_attach(self):
        """Reader cannot attach vif

        baremetal:node:vif:attach
        """
        try:
            resp = self.reader_client.vif_attach(
                self.node['uuid'], 'vifid')
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_vif_detach(self):
        """Reader cannot detach vif

        baremetal:node:vif:detach
        """
        try:
            resp, body = self.reader_client.vif_detach(
                self.node['uuid'], 'vifid')
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_can_traits_list(self):
        """Reader can list traits

        baremetal:node:traits:list
        """
        resp, body = self.reader_client.list_node_traits(self.node['uuid'])
        self.assertEqual(200, resp.status)

    def test_reader_cannot_traits_set(self):
        """Reader cannot set traits

        baremetal:node:traits:set
        """
        try:
            resp, body = self.reader_client.set_node_traits(
                self.node['uuid'], ['CUSTOM_TRAIT_A', 'CUSTOM_TRAIT_B'])
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_traits_delete(self):
        """Reader cannot delete traits

        baremetal:node:traits:delete
        """
        try:
            resp, body = self.reader_client.remove_node_traits(
                self.node['uuid'])
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_can_bios_get(self):
        """Reader can get bios settings

        baremetal:node:bios:get
        """
        resp, body = self.reader_client.list_node_bios_settings(
            self.node['uuid'])
        self.assertEqual(200, resp.status)

    def test_reader_cannot_disable_cleaning(self):
        """Reader cannot disable automated node cleaning

        baremetal:node:disable_cleaning
        """
        patch = [{'path': '/automated_clean', 'op': 'replace', 'value': False}]
        try:
            resp, body = self.reader_client.update_node(self.node['uuid'],
                                                        patch=patch)
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_history_get(self):
        """Reader cannot list history entries for a node

        baremetal:node:history:get
        """
        try:
            resp, body = self.reader_client.list_node_history(
                self.node['uuid'])
        except lib_exc.NotFound as e:
            resp = e.resp

        self.assertEqual(404, resp.status)

    def test_reader_cannot_vendor_passthru(self):
        """Reader cannot list vendor-specific extensions

        baremetal:node:vendor_passthru
        """
        try:
            resp, body = self.reader_client.list_vendor_passthru_methods(
                self.node['uuid'])
        except lib_exc.Forbidden as e:
            resp = e.resp

        self.assertEqual(403, resp.status)

    def test_reader_cannot_ipa_heartbeat(self):
        """Reader cannot heartbeat

        baremetal:node:ipa_heartbeat
        """
        # TODO(hjensas)
        pass



#
# Copyright 2017 Mirantis Inc.
#
# Copyright (c) 2022 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import os

import jsonschema
from jsonschema import exceptions as json_schema_exc
from oslo_log import log as logging
from tempest.common import utils
from tempest import config
from tempest.lib import decorators

from ironic_tempest_plugin import exceptions
from ironic_tempest_plugin.tests.scenario import \
    baremetal_standalone_manager as bsm

LOG = logging.getLogger(__name__)
CONF = config.CONF


class BaremetalCleaningIpmiWholedisk(
        bsm.BaremetalStandaloneScenarioTest):

    driver = 'ipmi'
    image_ref = CONF.baremetal.whole_disk_image_ref
    wholedisk_image = True
    delete_node = False
    deploy_interface = 'iscsi'
    api_microversion = '1.31'

    @classmethod
    def skip_checks(cls):
        super(BaremetalCleaningIpmiWholedisk, cls).skip_checks()
        if CONF.baremetal_feature_enabled.software_raid:
            raise cls.skipException("Cleaning is covered in the RAID test")

    @decorators.idempotent_id('065238db-1b6d-4d75-a9da-c240f8cbd956')
    @utils.services('image', 'network')
    def test_manual_cleaning(self):
        self.check_manual_partition_cleaning(self.node)


class SoftwareRaidIscsi(bsm.BaremetalStandaloneScenarioTest):

    if 'redfish' in CONF.baremetal.enabled_hardware_types:
        driver = 'redfish'
    else:
        driver = 'ipmi'
    image_ref = CONF.baremetal.whole_disk_image_ref
    wholedisk_image = True
    deploy_interface = 'iscsi'
    raid_interface = 'agent'
    api_microversion = '1.55'
    # Software RAID is always local boot
    boot_option = 'local'
    delete_node = False

    raid_config = {
        "logical_disks": [
            {
                "size_gb": "MAX",
                "raid_level": "1",
                "controller": "software"
            },
        ]
    }

    @classmethod
    def skip_checks(cls):
        super(SoftwareRaidIscsi, cls).skip_checks()
        if cls.driver == 'ipmi':
            raise cls.skipException("Testing with redfish driver")
        if not CONF.baremetal_feature_enabled.software_raid:
            raise cls.skipException("Software RAID feature is not enabled")

    @decorators.idempotent_id('7ecba4f7-98b8-4ea1-b95e-3ec399f46798')
    @utils.services('image', 'network')
    def test_software_raid(self):
        self.build_raid_and_verify_node(
            deploy_time=CONF.baremetal_feature_enabled.deploy_time_raid)
        # NOTE(TheJulia): tearing down/terminating the instance does not
        # remove the root device hint, so it is best for us to go ahead
        # and remove it before exiting the test.
        self.remove_root_device_hint()
        self.terminate_node(self.node['uuid'], force_delete=True)


class SoftwareRaidDirect(bsm.BaremetalStandaloneScenarioTest):

    if 'redfish' in CONF.baremetal.enabled_hardware_types:
        driver = 'redfish'
    else:
        driver = 'ipmi'
    image_ref = CONF.baremetal.whole_disk_image_ref
    wholedisk_image = True
    deploy_interface = 'direct'
    raid_interface = 'agent'
    api_microversion = '1.55'
    # Software RAID is always local boot
    boot_option = 'local'
    delete_node = False

    # TODO(dtantsur): more complex layout in this job
    raid_config = {
        "logical_disks": [
            {
                "size_gb": "MAX",
                "raid_level": "1",
                "controller": "software"
            },
        ]
    }

    @classmethod
    def skip_checks(cls):
        super(SoftwareRaidDirect, cls).skip_checks()
        if cls.driver == 'redfish':
            raise cls.skipException("Testing with ipmi driver")
        if not CONF.baremetal_feature_enabled.software_raid:
            raise cls.skipException("Software RAID feature is not enabled")

    @decorators.idempotent_id('125361ac-0eb3-4d79-8be2-a91936aa3f46')
    @utils.services('image', 'network')
    def test_software_raid(self):
        self.build_raid_and_verify_node(
            deploy_time=CONF.baremetal_feature_enabled.deploy_time_raid)
        # NOTE(TheJulia): tearing down/terminating the instance does not
        # remove the root device hint, so it is best for us to go ahead
        # and remove it before exiting the test.
        self.remove_root_device_hint()
        self.terminate_node(self.node['uuid'], force_delete=True)


class BaremetalIdracManagementCleaning(
        bsm.BaremetalStandaloneScenarioTest):

    mandatory_attr = ['driver', 'management_interface',
                      'power_interface']

    driver = 'idrac'
    delete_node = False
    # Minimum version for manual cleaning is 1.15 (# v1.15: Add ability to
    # do manual cleaning of nodes). The test cases clean up at the end by
    # detaching the VIF. Support for VIFs was introduced by version 1.28
    # (# v1.28: Add vifs subcontroller to node).
    api_microversion = '1.28'

    @decorators.idempotent_id('d085ff72-abef-4931-a5b0-06efd5f9a037')
    def test_reset_idrac(self):
        clean_steps = [
            {
                "interface": "management",
                "step": "reset_idrac"
            }
        ]
        self.manual_cleaning(self.node, clean_steps=clean_steps)

    @decorators.idempotent_id('9252ec6f-6b5b-447e-a323-c52775b88b4e')
    def test_clear_job_queue(self):
        clean_steps = [
            {
                "interface": "management",
                "step": "clear_job_queue"
            }
        ]
        self.manual_cleaning(self.node, clean_steps=clean_steps)

    @decorators.idempotent_id('7baeff52-7d6e-4dea-a48f-a85a6bfc9f62')
    def test_known_good_state(self):
        clean_steps = [
            {
                "interface": "management",
                "step": "known_good_state"
            }
        ]
        self.manual_cleaning(self.node, clean_steps=clean_steps)


class BaremetalIdracRedfishManagementCleaning(
        BaremetalIdracManagementCleaning):

    management_interface = 'idrac-redfish'
    power_interface = 'idrac-redfish'


class BaremetalIdracWSManManagementCleaning(
        BaremetalIdracManagementCleaning):

    management_interface = 'idrac-wsman'
    power_interface = 'idrac-wsman'


class BaremetalIdracRaidCleaning(bsm.BaremetalStandaloneScenarioTest):

    mandatory_attr = ['driver', 'raid_interface']
    image_ref = CONF.baremetal.whole_disk_image_ref
    wholedisk_image = True
    storage_inventory_info = None
    driver = 'idrac'
    api_microversion = '1.31'  # to set raid_interface
    delete_node = False

    @classmethod
    def skip_checks(cls):
        """Validates the storage information passed in file using JSON schema.

        :raises: skipException if,
            1) storage inventory path is not provided in tempest execution
            file.
            2) storage inventory file is not found on given path.
        :raises: RaidCleaningInventoryValidationFailed if,
            validation of the storage inventory fails.
        """
        super(BaremetalIdracRaidCleaning, cls).skip_checks()
        storage_inventory = CONF.baremetal.storage_inventory_file
        if not storage_inventory:
            raise cls.skipException("Storage inventory file path missing "
                                    "in tempest configuration file. "
                                    "Skipping Test case.")
        try:
            with open(storage_inventory, 'r') as storage_invent_fobj:
                cls.storage_inventory_info = json.load(storage_invent_fobj)
        except IOError:
            msg = ("Storage Inventory file %(inventory)s is not found. "
                   "Skipping Test Case." %
                   {'inventory': storage_inventory})
            raise cls.skipException(msg)
        storage_inventory_schema = os.path.join(os.path.dirname(
            __file__), 'storage_inventory_schema.json')
        with open(storage_inventory_schema, 'r') as storage_schema_fobj:
            schema = json.load(storage_schema_fobj)
        try:
            jsonschema.validate(cls.storage_inventory_info, schema)
        except json_schema_exc.ValidationError as e:
            error_msg = ("Storage Inventory validation error: %(error)s " %
                         {'error': e})
            raise exceptions.RaidCleaningInventoryValidationFailed(error_msg)

    def _validate_raid_type_and_drives_count(self, raid_type,
                                             minimum_drives_required):
        for controller in (self.storage_inventory_info[
                           'storage_inventory']['controllers']):
            supported_raid_types = controller['supported_raid_types']
            physical_disks = [pdisk['id'] for pdisk in (
                controller['drives'])]
            if raid_type in supported_raid_types and (
                    minimum_drives_required <= len(physical_disks)):
                return controller
        error_msg = ("No Controller present in storage inventory which "
                     "supports RAID type %(raid_type)s "
                     "and has at least %(disk_count)s drives." %
                     {'raid_type': raid_type,
                      'disk_count': minimum_drives_required})
        raise exceptions.RaidCleaningInventoryValidationFailed(error_msg)

    @decorators.idempotent_id('8a908a3c-f2af-48fb-8553-9163715aa403')
    @utils.services('image', 'network')
    def test_hardware_raid(self):
        controller = self._validate_raid_type_and_drives_count(
            raid_type='1', minimum_drives_required=2)
        raid_config = {
            "logical_disks": [
                {
                    "size_gb": 40,
                    "raid_level": "1",
                    "controller": controller['id']
                }
            ]
        }
        self.build_raid_and_verify_node(
            config=raid_config,
            deploy_time=CONF.baremetal_feature_enabled.deploy_time_raid,
            erase_device_metadata=False)
        self.remove_root_device_hint()
        self.terminate_node(self.node['uuid'], force_delete=True)

    @decorators.idempotent_id('92fe534d-77f1-422d-84e4-e30fe9e3d928')
    @utils.services('image', 'network')
    def test_raid_cleaning_max_size_raid_10(self):
        controller = self._validate_raid_type_and_drives_count(
            raid_type='1+0', minimum_drives_required=4)
        physical_disks = [pdisk['id'] for pdisk in (
            controller['drives'])]
        raid_config = {
            "logical_disks": [
                {
                    "size_gb": "MAX",
                    "raid_level": "1+0",
                    "controller": controller['id'],
                    "physical_disks": physical_disks
                }
            ]
        }
        self.build_raid_and_verify_node(
            config=raid_config,
            deploy_time=CONF.baremetal_feature_enabled.deploy_time_raid,
            erase_device_metadata=False)
        self.remove_root_device_hint()
        self.terminate_node(self.node['uuid'], force_delete=True)


class BaremetalIdracRedfishRaidCleaning(
        BaremetalIdracRaidCleaning):
    raid_interface = 'idrac-redfish'


class BaremetalIdracWSManRaidCleaning(
        BaremetalIdracRaidCleaning):
    raid_interface = 'idrac-wsman'


class BaremetalRedfishFirmwareUpdate(bsm.BaremetalStandaloneScenarioTest):

    api_microversion = '1.68'  # to support redfish firmware update
    driver = 'redfish'
    delete_node = False
    image_ref = CONF.baremetal.whole_disk_image_ref
    wholedisk_image = True

    @classmethod
    def skip_checks(cls):
        super(BaremetalRedfishFirmwareUpdate, cls).skip_checks()
        if not CONF.baremetal.firmware_image_url:
            raise cls.skipException("Firmware image URL is not "
                                    "provided. Skipping test case.")
        if not CONF.baremetal.firmware_image_checksum:
            raise cls.skipException("Firmware image SHA1 checksum is not "
                                    "provided. Skipping test case.")

    def _firmware_update(self, fw_image_url, fw_image_checksum):
        steps = [
            {
                "interface": "management",
                "step": "update_firmware",
                "args": {
                    "firmware_images": [
                        {
                            "url": fw_image_url,
                            "checksum": fw_image_checksum,
                            "wait": 300
                        }
                    ]
                }
            }
        ]
        self.manual_cleaning(self.node, clean_steps=steps)

    @utils.services('network')
    @decorators.idempotent_id('360e0c0e-3c17-4d2e-b052-55a932c1a4c7')
    def test_firmware_update(self):
        # WARNING: Removing power from a server while it is in the process of
        # updating firmware may result in devices in the server, or the server
        # itself becoming inoperable.
        # Execution of firmware test case needs careful execution as it
        # changes state of server, may result in break down of server on
        # interruption. As it deals with firmware of component, make sure
        # to provide proper image url path while testing with correct SHA1
        # checksum of image.
        self._firmware_update(CONF.baremetal.firmware_image_url,
                              CONF.baremetal.firmware_image_checksum)
        if (CONF.baremetal.firmware_rollback_image_url
                and CONF.baremetal.firmware_rollback_image_checksum):
            self.addCleanup(self._firmware_update,
                            CONF.baremetal.firmware_rollback_image_url,
                            CONF.baremetal.firmware_rollback_image_checksum)


class BaremetalIdracRedfishFirmwareUpdate(BaremetalRedfishFirmwareUpdate):

    driver = 'idrac'
    boot_interface = 'ipxe'
    management_interface = 'idrac-redfish'
    power_interface = 'idrac-redfish'


class BaremetalIdracRedfishConfigurationMolds(
        bsm.BaremetalStandaloneScenarioTest):

    api_microversion = '1.72'  # to support configuration molds functionality
    delete_node = False
    image_ref = CONF.baremetal.whole_disk_image_ref
    driver = 'idrac'
    boot_interface = 'ipxe'
    management_interface = 'idrac-redfish'
    power_interface = 'idrac-redfish'
    wholedisk_image = True

    @utils.services('network')
    @decorators.idempotent_id('69386cf6-bbf2-451e-b927-f68c66075f02')
    def test_configuration_molds_export(self):
        if not CONF.baremetal.export_location:
            raise self.skipException("Export configuration location path "
                                     "is not provided. Skipping test case. "
                                     "Make sure to provide correct "
                                     "configuration path in which "
                                     "configuration would be exported. "
                                     "In case of Swift object storage, "
                                     "make to sure to provide proper "
                                     "container path.")

        steps = [
            {
                "interface": "management",
                "step": "export_configuration",
                "args": {
                    "export_configuration_location":
                        CONF.baremetal.export_location
                }
            }
        ]
        self.manual_cleaning(self.node, clean_steps=steps)

    @utils.services('network')
    @decorators.idempotent_id('cad15719-c293-4338-8fa9-986ef02b4682')
    def test_configuration_molds_import(self):
        if not CONF.baremetal.import_location:
            raise self.skipException("Import configuration JSON file location "
                                     "is not provided. Make sure to provide "
                                     "correct configuration file with proper "
                                     "configuration which needs to be tested "
                                     "during execution of this test. "
                                     "Skipping test case.")
        steps = [
            {
                "interface": "management",
                "step": "import_configuration",
                "args": {
                    "import_configuration_location":
                        CONF.baremetal.import_location
                }
            }
        ]
        self.manual_cleaning(self.node, clean_steps=steps)
        if CONF.baremetal.rollback_import_location:
            rollback_steps = [
                {
                    "interface": "management",
                    "step": "import_configuration",
                    "args": {
                        "import_configuration_location":
                            CONF.baremetal.rollback_import_location
                    }
                }
            ]
            self.addCleanup(self.manual_cleaning, self.node,
                            clean_steps=rollback_steps)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest import config
from tempest.lib import decorators

from ironic_tempest_plugin.tests.api import base


CONF = config.CONF


class TestAddShardsToNode(base.BaseBaremetalTest):
    """Tests for baremetal shards."""

    min_microversion = '1.82'

    def setUp(self):
        super(TestAddShardsToNode, self).setUp()
        # set a minimum API version
        _, self.chassis = self.create_chassis()

    @decorators.idempotent_id('6f1e241d-4386-4730-b9ff-28c6a3dcad31')
    def test_add_shard_to_node_at_create(self):
        shard = 'at-create'

        _, body = self.create_node(self.chassis['uuid'], shard=shard)
        self.assertEqual(shard, body['shard'])

    @decorators.idempotent_id('2eb91d29-e0a5-472b-aeb8-ef6d98eb0f3c')
    def test_add_shard_to_node_post_create(self):
        shard = 'post-create'

        _, node = self.create_node(self.chassis['uuid'])
        _, before = self.client.show_node(node['uuid'])
        self.assertIsNone(before['shard'])

        self.client.update_node(node['uuid'], shard=shard)

        _, after = self.client.show_node(node['uuid'])
        self.assertEqual(shard, after['shard'])


class TestNodeShardQueries(base.BaseBaremetalTest):
    """Tests for baremetal shards."""

    min_microversion = '1.82'

    def setUp(self):
        super(TestNodeShardQueries, self).setUp()
        _, self.chassis = self.create_chassis()
        _, bad_node = self.create_node(self.chassis['uuid'], shard='bad')
        _, none_node = self.create_node(self.chassis['uuid'])  # shard=None
        self.bad_node_id = bad_node['uuid']
        self.none_node_id = none_node['uuid']

    def _setup_nodes(self, good_shard, num=2):
        good_node_ids = []
        for i in range(num):
            _, node = self.create_node(self.chassis['uuid'], shard=good_shard)
            good_node_ids.append(node['uuid'])

        return good_node_ids

    @decorators.idempotent_id('df74c989-6972-4104-a8d6-bd8e8d811353')
    def test_show_all_nodes(self):
        """Validate unfiltered API query will return nodes with a shard."""
        shard = "oneshardtest"
        good_node_ids = self._setup_nodes(shard)

        _, fetched_nodes = self.client.list_nodes()
        fetched_node_ids = [node['uuid'] for node in fetched_nodes['nodes']]

        for node_id in good_node_ids:
            self.assertIn(node_id, fetched_node_ids)

    @decorators.idempotent_id('6f1e241d-4386-4730-b9ff-28c6a3dcad31')
    def test_only_show_requested_shard(self):
        """Validate filtering nodes by shard."""
        shard = "oneshardtest"
        good_node_ids = self._setup_nodes(shard)

        _, fetched_nodes = self.client.list_nodes(shard=shard)
        fetched_node_ids = [node['uuid'] for node in fetched_nodes['nodes']]

        self.assertItemsEqual(good_node_ids, fetched_node_ids)

    @decorators.idempotent_id('6f1e241d-4386-4730-b9ff-28c6a3dcad31')
    def test_only_show_multiple_requested_shards(self):
        """Validate filtering nodes by multiple shards."""
        shard = "multishardtest"
        second_shard = "multishardtest2"
        good_node_ids = self._setup_nodes(shard)
        _, second_shard_node = self.create_node(
            self.chassis['uuid'], shard=second_shard)
        good_node_ids.append(second_shard_node['uuid'])

        _, fetched_nodes = self.client.list_nodes(
            shard=','.join([shard, second_shard]))
        fetched_node_ids = [node['uuid'] for node in fetched_nodes['nodes']]

        self.assertItemsEqual(good_node_ids, fetched_node_ids)
        self.assertNotIn(self.bad_node_id, fetched_node_ids)
        self.assertNotIn(self.none_node_id, fetched_node_ids)

    @decorators.idempotent_id('f7a2eeb7-d16e-480c-b698-3448491c73a1')
    def test_show_sharded_nodes(self):
        _, fetched_nodes = self.client.list_nodes(sharded=True)
        fetched_node_ids = [node['uuid'] for node in fetched_nodes['nodes']]

        # NOTE(JayF): All other nodes under test are sharded
        self.assertNotIn(self.none_node_id, fetched_node_ids)

    @decorators.idempotent_id('f7a2eeb7-d16e-480c-b698-3448491c73a1')
    def test_show_unsharded_nodes(self):
        _, fetched_nodes = self.client.list_nodes(sharded=False)
        fetched_node_ids = [node['uuid'] for node in fetched_nodes['nodes']]

        self.assertIn(self.none_node_id, fetched_node_ids)
        self.assertNotIn(self.bad_node_id, fetched_node_ids)


class TestGetAllShards(base.BaseBaremetalTest):
    """Tests for baremetal shards."""

    min_microversion = '1.82'

    def setUp(self):
        super(TestGetAllShards, self).setUp()
        _, self.chassis = self.create_chassis()
        self.shards = ["shard1", "shard2", "shard3"]
        self.node_ids = []
        for shard in self.shards:
            _, node = self.create_node(self.chassis['uuid'], shard=shard)
            self.node_ids.append(node['uuid'])

    @decorators.idempotent_id('fc786196-63c7-4e0d-bd14-3e478d4d1e3e')
    def test_get_all_shards(self):
        _, fetched_shards = self.client.get_shards()
        fetched_shards = [shard['name'] for shard in fetched_shards['shards']]

        self.assertItemsEqual(self.shards, fetched_shards)



#    Copyright (c) 2022 Dell Inc. or its subsidiaries.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest import config
from tempest.lib import decorators

from ironic_tempest_plugin.tests.scenario import \
    baremetal_standalone_manager as bsm

CONF = config.CONF


class BaremetalInspectBase:

    mandatory_attr = ['driver', 'inspect_interface']
    # (# v1.31: Support for updating inspect_interface).
    api_microversion = '1.31'
    delete_node = False
    wait_provisioning_state_interval = 1

    def _verify_node_inspection_data(self, node):
        self.assertIn(node['properties']['cpu_arch'], ['x86_64', 'aarch64'])

    @decorators.idempotent_id('47ea4487-4720-43e8-a024-53ae82f8c264')
    def test_baremetal_inspect(self):
        """This test case follows this set of operations:

            * Sets nodes to manageable state
            * Inspects nodes
            * Verifies inspection data
            * Sets node to available state
        """
        self.baremetal_client.set_node_provision_state(self.node['uuid'],
                                                       'manage')
        _, node = self.baremetal_client.show_node(self.node['uuid'])
        if 'cpu_arch' in node['properties']:
            new_properties = node['properties'].copy()
            new_properties.pop('cpu_arch')
            self.baremetal_client.update_node(self.node['uuid'],
                                              properties=new_properties)

        self.baremetal_client.set_node_provision_state(self.node['uuid'],
                                                       'inspect')
        self.wait_provisioning_state(
            self.node['uuid'], 'manageable',
            timeout=CONF.baremetal.inspect_timeout,
            interval=self.wait_provisioning_state_interval)

        _, node = self.baremetal_client.show_node(self.node['uuid'])
        self._verify_node_inspection_data(node)

        self.baremetal_client.set_node_provision_state(self.node['uuid'],
                                                       'provide')
        self.wait_provisioning_state(self.node['uuid'], 'available')


class BaremetalRedfishAgentInspect(BaremetalInspectBase,
                                   bsm.BaremetalStandaloneScenarioTest):
    driver = 'redfish'
    inspect_interface = 'agent'
    wait_provisioning_state_interval = 15
    # 1.81 adds support for inventory API
    api_microversion = '1.81'

    def _verify_node_inspection_data(self, node):
        super()._verify_node_inspection_data(node)
        inspection_data = self.baremetal_client.show_inventory(
            self.node['uuid'])
        self.assertEqual({'inventory', 'plugin_data'}, set(inspection_data))

        # Inventory sanity check
        inventory = inspection_data['inventory']
        self.assertGreater(inventory['cpu']['count'], 0)
        self.assertGreater(inventory['memory']['physical_mb'], 256)
        self.assertGreater(len(inventory['disks']), 0)
        self.assertGreater(len(inventory['interfaces']), 0)

    @decorators.idempotent_id('82670a85-49b7-4d28-b8a8-6e18b0d4c8d1')
    def test_inspect_abort(self):
        _, node = self.baremetal_client.show_node(self.node['uuid'])
        current_state = node['provision_state']

        def cleanup():
            nonlocal current_state

            if current_state == 'inspect failed':
                self.baremetal_client.set_node_provision_state(
                    self.node['uuid'], 'manage')
                self.wait_provisioning_state(self.node['uuid'], 'manageable')
                current_state = 'manageable'

            if current_state == 'manageable':
                self.baremetal_client.set_node_provision_state(
                    self.node['uuid'], 'provide')
                self.wait_provisioning_state(self.node['uuid'], 'available')

        self.addCleanup(cleanup)

        if current_state != 'manageable':
            self.baremetal_client.set_node_provision_state(self.node['uuid'],
                                                           'manage')
        current_state = 'manageable'

        self.baremetal_client.set_node_provision_state(self.node['uuid'],
                                                       'inspect')
        self.wait_provisioning_state(
            self.node['uuid'], 'inspect wait',
            timeout=300, interval=5)

        self.baremetal_client.set_node_provision_state(self.node['uuid'],
                                                       'abort')
        self.wait_provisioning_state(
            self.node['uuid'], 'inspect failed',
            timeout=60, interval=1)
        current_state = 'inspect failed'


class BaremetalIdracInspect(BaremetalInspectBase,
                            bsm.BaremetalStandaloneScenarioTest):
    driver = 'idrac'

    def _verify_node_inspection_data(self, node):
        super()._verify_node_inspection_data(node)
        self.assertGreater(int(node['properties']['memory_mb']), 0)
        self.assertGreater(int(node['properties']['cpus']), 0)
        self.assertGreater(int(node['properties']['local_gb']), 0)


class BaremetalIdracRedfishInspect(BaremetalIdracInspect):
    inspect_interface = 'idrac-redfish'


class BaremetalIdracWSManInspect(BaremetalIdracInspect):
    inspect_interface = 'idrac-wsman'



#
# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging
from tempest.common import utils
from tempest.common import waiters
from tempest import config
from tempest.lib.common import api_version_request
from tempest.lib import decorators

from ironic_tempest_plugin.tests.scenario import baremetal_manager

LOG = logging.getLogger(__name__)
CONF = config.CONF


class BaremetalBasicOps(baremetal_manager.BaremetalScenarioTest):
    """This smoke test tests an Ironic driver.

    It follows this basic set of operations:
        * Creates a keypair
        * Boots an instance using the keypair
        * Monitors the associated Ironic node for power and
          expected state transitions
        * Validates Ironic node's port data has been properly updated
        * Validates Ironic node's resource class and traits have been honoured
        * Verifies SSH connectivity using created keypair via fixed IP
        * Associates a floating ip
        * Verifies SSH connectivity using created keypair via floating IP
        * Verifies instance rebuild with ephemeral partition preservation
        * Deletes instance
        * Monitors the associated Ironic node for power and
          expected state transitions
    """

    credentials = ['primary', 'admin', 'system_admin']

    TEST_RESCUE_MODE = False
    image_ref = None
    wholedisk_image = None
    auto_lease = False

    @classmethod
    def skip_checks(cls):
        super(BaremetalBasicOps, cls).skip_checks()

        # If default rescue interface is configured to test the rescue
        # feature, then skips this test and let the test derived class
        # to be executed.
        rescue_if = CONF.baremetal.default_rescue_interface
        if cls.TEST_RESCUE_MODE:
            if not rescue_if or rescue_if == 'no-rescue':
                msg = 'Node rescue interface is not enabled.'
                raise cls.skipException(msg)
        else:
            if rescue_if and rescue_if != 'no-rescue':
                msg = ('Node rescue interface is enabled, but %s class '
                       'cannot test rescue operations.' % cls.__name__)
                raise cls.skipException(msg)

    @staticmethod
    def _is_version_supported(version):
        """Return whether an API microversion is supported."""
        min_version = api_version_request.APIVersionRequest(
            CONF.baremetal.min_microversion)
        max_version = api_version_request.APIVersionRequest(
            CONF.baremetal.max_microversion)
        version = api_version_request.APIVersionRequest(version)
        return min_version <= version <= max_version

    def rebuild_instance(self, preserve_ephemeral=False):
        self.rebuild_server(server_id=self.instance['id'],
                            preserve_ephemeral=preserve_ephemeral,
                            wait=False)

        node = self.get_node(instance_id=self.instance['id'])

        # We should remain on the same node
        self.assertEqual(self.node['uuid'], node['uuid'])
        self.node = node

        waiters.wait_for_server_status(
            self.servers_client,
            server_id=self.instance['id'],
            status='REBUILD',
            ready_wait=False)
        waiters.wait_for_server_status(
            self.servers_client,
            server_id=self.instance['id'],
            status='ACTIVE')

    def verify_partition(self, client, label, mount, gib_size):
        """Verify a labeled partition's mount point and size."""
        LOG.info("Looking for partition %s mounted on %s", label, mount)

        # Validate we have a device with the given partition label
        cmd = "/sbin/blkid -c /dev/null -l -o device -t LABEL=%s" % label
        device = client.exec_command(cmd).rstrip('\n')
        LOG.debug("Partition device is %s", device)
        self.assertNotEqual('', device)

        # Validate the mount point for the device
        cmd = "mount | grep -w '%s' | cut -d' ' -f3" % device
        actual_mount = client.exec_command(cmd).rstrip('\n')
        LOG.debug("Partition mount point is %s", actual_mount)
        self.assertEqual(actual_mount, mount)

        # Validate the partition size matches what we expect
        numbers = '0123456789'
        devnum = device.replace('/dev/', '')
        cmd = "cat /sys/block/%s/%s/size" % (devnum.rstrip(numbers), devnum)
        num_bytes = client.exec_command(cmd).rstrip('\n')
        num_bytes = int(num_bytes) * 512
        actual_gib_size = num_bytes / (1024 * 1024 * 1024)
        LOG.debug("Partition size is %d GiB", actual_gib_size)
        self.assertEqual(actual_gib_size, gib_size)

    def get_flavor_ephemeral_size(self):
        """Returns size of the ephemeral partition in GiB."""
        f_id = self.instance['flavor']['id']
        flavor = self.flavors_client.show_flavor(f_id)['flavor']
        ephemeral = flavor.get('OS-FLV-EXT-DATA:ephemeral')
        if not ephemeral or ephemeral == 'N/A':
            return None
        return int(ephemeral)

    def validate_ports(self):
        node_uuid = self.node['uuid']
        vifs = self.get_node_vifs(node_uuid)

        ir_ports = self.get_ports(node_uuid)
        ir_ports_addresses = [x['address'] for x in ir_ports]
        for vif in vifs:
            n_port_id = vif['id']
            body = self.ports_client.show_port(n_port_id)
            n_port = body['port']
            self.assertEqual(n_port['device_id'], self.instance['id'])
            self.assertIn(n_port['mac_address'], ir_ports_addresses)

    def validate_scheduling(self):
        """Validate scheduling attributes of the node against the flavor.

        Validates the resource class and traits requested by the flavor against
        those set on the node. Does not assume that resource classes and traits
        are in use.
        """
        node = self.get_node(instance_id=self.instance['id'],
                             api_version='1.37')
        f_id = self.instance['flavor']['id']
        extra_specs = self.flavors_client.list_flavor_extra_specs(f_id)
        extra_specs = extra_specs['extra_specs']

        # Pull the requested resource class and traits from the flavor.
        resource_class = None
        traits = set()
        for key, value in extra_specs.items():
            if key.startswith('resources:CUSTOM_') and value == '1':
                resource_class = key.partition(':')[2]
            if key.startswith('trait:') and value == 'required':
                trait = key.partition(':')[2]
                traits.add(trait)

        # Validate requested resource class and traits against the node.
        if resource_class is not None:
            # The resource class in ironic may be lower case, and must omit the
            # CUSTOM_ prefix. Normalise it.
            node_resource_class = node['resource_class']
            node_resource_class = node_resource_class.upper()
            node_resource_class = node_resource_class.translate(
                str.maketrans(" -.:", "____", "!@#$%^&*()+=/\\?<>|\"'")
            )
            node_resource_class = 'CUSTOM_' + node_resource_class
            self.assertEqual(resource_class, node_resource_class)

        if 'traits' in node and traits:
            self.assertIn('traits', node['instance_info'])
            # All flavor traits should be added as instance traits.
            self.assertEqual(traits, set(node['instance_info']['traits']))
            # Flavor traits should be a subset of node traits.
            self.assertTrue(traits.issubset(set(node['traits'])))

    def validate_image(self):
        iinfo = self.node['instance_info']
        if self.wholedisk_image is not None and self.wholedisk_image:
            # If None, we have nothing to do here. If False, we don't
            # want to fall into this either.
            self.assertNotIn('kernel', iinfo)
            self.assertNotIn('ramdisk', iinfo)
        else:
            if 'image_type' in iinfo:
                self.assertEqual('partition', iinfo['image_type'])
            else:
                self.assertTrue(iinfo['kernel'])
                self.assertTrue(iinfo['ramdisk'])
            self.assertGreater(int(iinfo['root_gb']), 0)

    def validate_uefi(self, client):
        efi_dir = '/sys/firmware/efi'
        success_string = "Found " + efi_dir
        cmd = 'test -d {dir} && echo "Found {dir}" ||'\
              ' echo "{dir} not found"'.format(dir=efi_dir)
        output = client.exec_command(cmd).rstrip()
        self.assertEqual(success_string, output)

    def validate_lessee(self):
        iinfo = self.node.get('instance_info')
        dii = self.node.get('driver_internal_info', {})
        if 'automatic_lessee' in dii and iinfo:
            # NOTE(JayF): This item not being in instance_info tells us we
            # set the lessee.
            self.assertEqual(iinfo['project_id'], self.node['lessee'])

    def baremetal_server_ops(self):
        self.add_keypair()
        self.instance, self.node = self.boot_instance(image_id=self.image_ref)
        self.validate_image()
        self.validate_ports()
        self.validate_scheduling()
        self.validate_lessee()
        ip_address = self.get_server_ip(self.instance)
        self.check_vm_connectivity(ip_address=ip_address,
                                   private_key=self.keypair['private_key'],
                                   server=self.instance)
        vm_client = self.get_remote_client(ip_address, server=self.instance)

        # We expect the ephemeral partition to be mounted on /mnt and to have
        # the same size as our flavor definition.
        eph_size = self.get_flavor_ephemeral_size()
        if eph_size and not self.wholedisk_image:
            self.verify_partition(vm_client, 'ephemeral0', '/mnt', eph_size)
            # Create the test file
            self.create_timestamp(ip_address,
                                  private_key=self.keypair['private_key'],
                                  server=self.instance)

        if CONF.baremetal.boot_mode == "uefi":
            self.validate_uefi(vm_client)

        # Test rescue mode
        if self.TEST_RESCUE_MODE:
            self.rescue_instance(self.instance, self.node, ip_address)
            self.unrescue_instance(self.instance, self.node, ip_address)

        self.terminate_instance(self.instance)

    @decorators.idempotent_id('549173a5-38ec-42bb-b0e2-c8b9f4a08943')
    @utils.services('compute', 'image', 'network')
    def test_baremetal_server_ops_partition_image(self):
        # NOTE(dtantsur): cirros partition images don't have grub, we cannot
        # use local boot on BIOS with them.
        if (CONF.baremetal.partition_netboot
                and CONF.baremetal.default_boot_option == 'local'):
            raise self.skipException(
                "Cannot test partition images with local boot on cirros")

        self.image_ref = CONF.baremetal.partition_image_ref
        self.wholedisk_image = False
        self.baremetal_server_ops()

    @decorators.idempotent_id('d7d48aa1-0395-4f31-a908-60969adc4322')
    @utils.services('compute', 'image', 'network')
    def test_baremetal_server_ops_wholedisk_image(self):
        self.image_ref = CONF.baremetal.whole_disk_image_ref
        self.wholedisk_image = True
        self.auto_lease = True
        self.baremetal_server_ops()


class BaremetalBasicOpsAndRescue(BaremetalBasicOps):
    """This test includes rescue/unrescue ops."""

    TEST_RESCUE_MODE = True



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest import clients
from tempest.common import credentials_factory as common_creds
from tempest import config

from ironic_tempest_plugin.services.baremetal import base


CONF = config.CONF


class Manager(clients.Manager):
    def __init__(self,
                 credentials=None,
                 api_microversions=None):
        if not credentials:
            credentials = common_creds.get_configured_admin_credentials()
        super(Manager, self).__init__(credentials)
        self.introspection_client = BaremetalIntrospectionClient(
            self.auth_provider,
            CONF.baremetal_introspection.catalog_type,
            CONF.identity.region,
            endpoint_type=CONF.baremetal_introspection.endpoint_type)


class BaremetalIntrospectionClient(base.BaremetalClient):
    """Base Tempest REST client for Ironic Inspector API v1."""
    version = '1'
    uri_prefix = 'v1'

    @base.handle_errors
    def purge_rules(self):
        """Purge all existing rules."""
        return self._delete_request('rules', uuid=None)

    @base.handle_errors
    def create_rules(self, rules):
        """Create introspection rules."""
        if not isinstance(rules, list):
            rules = [rules]
        for rule in rules:
            self._create_request('rules', rule)

    @base.handle_errors
    def get_status(self, uuid):
        """Get introspection status for a node."""
        return self._show_request('introspection', uuid=uuid)

    @base.handle_errors
    def get_data(self, uuid):
        """Get introspection data for a node."""
        return self._show_request('introspection', uuid=uuid,
                                  uri='/%s/introspection/%s/data' %
                                      (self.uri_prefix, uuid))

    @base.handle_errors
    def start_introspection(self, uuid):
        """Start introspection for a node."""
        resp, _body = self.post(url=('/%s/introspection/%s' %
                                     (self.uri_prefix, uuid)),
                                body=None)
        self.expected_success(202, resp.status)

        return resp

    @base.handle_errors
    def abort_introspection(self, uuid):
        """Abort introspection for a node."""
        resp, _body = self.post(url=('/%s/introspection/%s/abort' %
                                     (self.uri_prefix, uuid)),
                                body=None)
        self.expected_success(202, resp.status)

        return resp
